/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// models.c -- model loading and caching

// models are the only shared resource between a client and server running
// on the same machine.

#include "gl_state.h"

model_t		*loadmodel;
char		loadname[32];	// for hunk tags

void MH_WorldVolumeCarve (model_t *mod);
void Mod_LoadSpriteModel (model_t *mod, void *buffer);
void Mod_LoadBrushModel (model_t *mod, void *buffer);
void Mod_LoadAliasModel (model_t *mod, void *buffer);
model_t *Mod_LoadModel (model_t *mod, qboolean crash);

#define MAX_MOD_BUFFER  1024
#define	MAX_MOD_KNOWN	512
model_t			mod_known[MAX_MOD_KNOWN];
byte			mod_novis[MAX_MAP_LEAFS/8];
int				mod_numknown;
int				mod_filesize;

/*
===============
Mod_Init
===============
*/
void Mod_Init (void)
{
    memset (mod_novis, 0xff, sizeof(mod_novis));
}

/*
===============
Mod_PointInLeaf
===============
*/
mleaf_t *Mod_PointInLeaf (vec3_t p, model_t *model)
{
    mnode_t		*node;
    float		d;

    if (!model || !model->nodes)
    {
        Host_Error ("Mod_PointInLeaf: bad model");
    }
    node = model->nodes;

    while (1)
    {
        if (node->contents < 0)
        {
            return (mleaf_t *) node;
        }

        switch (node->plane->type)
        {
        case PLANE_X:
            d = p[0] - node->plane->dist;
            break;
        case PLANE_Y:
            d = p[1] - node->plane->dist;
            break;
        case PLANE_Z:
            d = p[2] - node->plane->dist;
            break;
        default:
            d = DotProduct (p, node->plane->normal) - node->plane->dist;
            break;
        }

        if (d > 0)
        {
            node = node->children[0];
        }
        else
        {
            node = node->children[1];
        }
    }

    // never reached
    return NULL;
}

/*
===================
Mod_DecompressVis
===================
*/
byte *Mod_DecompressVis (byte *in, model_t *model)
{
    static byte	decompressed[MAX_MAP_LEAFS/8];
    int			c;
    byte		*out;
    int			row;

    memset (decompressed, 0, sizeof (decompressed));
    row = (model->numleafs + 7) >> 3;
    out = decompressed;

    if (!in || r_novis.value == 2)
    {
        // no vis info, so make all visible
        while (row)
        {
            *out++ = 0xff;
            row--;
        }
        return decompressed;
    }

    do
    {
        if (*in)
        {
            *out++ = *in++;
            continue;
        }
        c = in[1];
        in += 2;

        while (c)
        {
            *out++ = 0;
            c--;
        }
    }
    while (out - decompressed < row);

    return decompressed;
}

/*
===================
Mod_LeafPVS
===================
*/
byte *Mod_LeafPVS (mleaf_t *leaf, model_t *model)
{
    // leaf 0 is the solid leaf
    if (leaf == model->leafs)
    {
        return mod_novis;
    }

    // no visdata
    if (!leaf->decompressed_vis)
    {
        return mod_novis;
    }

    // return the decompressed vis
    return leaf->decompressed_vis;
}

/*
===================
Mod_ClearAll
===================
*/
void Mod_ClearAll (void)
{
    int			i;
    model_t		*mod;

    // reload all models at map start
    for (i = 0 ,mod = mod_known ; i < mod_numknown ; i++, mod++)
    {
        if (mod->loaded && mod->name[0])
        {
            mod->loaded = false;
        }
    }
}

/*
==================
Mod_FindName
==================
*/
model_t *Mod_FindName (char *name)
{
    int		i;
    model_t	*mod;

    if (!name[0])
    {
        Sys_Error ("Mod_FindName: NULL name");
    }

    // search the currently loaded models
    for (i = 0, mod = mod_known; i < mod_numknown; i++, mod++)
    {
        if (!strcasecmp (mod->name, name)) break;
    }

    if (i == mod_numknown)
    {
        int len;

        if (mod_numknown == MAX_MOD_KNOWN)
        {
            Host_Error ("Mod_FindName: mod_numknown == MAX_MOD_KNOWN (%d)", MAX_MOD_KNOWN);
        }
        len = (int) strlen (name);

        if (len >= (int) sizeof(mod->name))
        {
            Host_Error ("Mod_FindName: name '%s' too long (%d, max = %d)", name, len, sizeof(mod->name) - 1);
        }
        strlcpy_s (mod->name, name);
        mod_numknown++;
    }
    return mod;
}

/*
==================
Mod_LoadModel

Loads a model into the cache
==================
*/
model_t *Mod_LoadModel (model_t *mod, qboolean crash)
{
    unsigned int *buf;

    // doesn't need to be loaded
    if (mod->loaded)
    {
        return mod;
    }

    // load the file into the temp hunk (it will be transferred to permanent
    // memory in it's correct memory format as part of the loading process)
    if (mod->name[0] != '/')
    {
        buf = (unsigned int *)COM_LoadTempFile (mod->name);
    }
    else
    {
        buf = (unsigned int *)COM_LoadTempFile ((mod->name + 1));
    }

    if (!buf)
    {
        // If last added mod_known, get rid of it
        if (mod == &mod_known[mod_numknown - 1])
        {
            --mod_numknown;
        }

        if (crash)
        {
            Host_Error ("Mod_LoadModel: %s not found", mod->name);
        }
        return NULL;
    }

    // Save off
    mod_filesize = com_filesize;

    // allocate a new model
    COM_FileBase (mod->name, loadname);

    loadmodel = mod;

    // its loaded
    mod->loaded = true;

    // call the apropriate loader
    switch (LittleLong (*(unsigned int *)buf))
    {
    case IDPOLYHEADER:
        Mod_LoadAliasModel (mod, buf);
        break;

    case IDSPRITEHEADER:
        Mod_LoadSpriteModel (mod, buf);
        break;

    default:
        Mod_LoadBrushModel (mod, buf);
        break;
    }
    return mod;
}

/*
==================
Mod_ForName

Loads in a model for the given name
==================
*/
model_t *Mod_ForName (char *name, qboolean crash)
{
    model_t	*mod;

    mod = Mod_FindName (name);

    return Mod_LoadModel (mod, crash);
}


/*
===============================================================================

					BRUSHMODEL LOADING

===============================================================================
*/

static int  MissFrameNo;
static char MissFrameName[16+1];
static byte	*mod_base;

/*
=================
Mod_WithinBounds
=================
*/
static qboolean Mod_WithinBounds (int SItem, int Items, int Limit, int *ErrItem)
{
    int EItem = 0;

    if (SItem < 0)
    {
        EItem = SItem;
    }
    else if (Items < 0)
    {
        EItem = Items;
    }
    else if (SItem + Items < 0 || SItem + Items > Limit)
    {
        EItem = SItem + Items;
    }

    if (ErrItem)
    {
        *ErrItem = EItem;
    }
    return (EItem == 0) ? true : false;
}

/*
=================
Mod_ChkAllFrames
=================
*/
static qboolean Mod_ChkAllFrames (texture_t *tx, texture_t *anims[], int max, qboolean Alternate)
{
    int i, frames = 0;

    for (i=0 ; i<max ; i++)
    {
        if (anims[i])
        {
            ++frames;
        }
        else
        {
            // Avoid unnecessary warnings
            if (i != MissFrameNo || strcmp(tx->name, MissFrameName))
            {
                MissFrameNo = i;
                strlcpy_s (MissFrameName, tx->name);
                Con_SafePrintf ("\002Mod_LoadTextures: ");
                Con_Printf ("missing %s frame %i of '%s'\n", Alternate ? "alternate" : "", i, tx->name);
            }
        }
    }
    return (frames == max) ? true : false;
}

/*
=================
Mod_ChkLumpSize
=================
*/
static void Mod_ChkLumpSize (char *Function, lump_t *l, int Size)
{
    if (l->filelen % Size)
    {
        // this is an error !!!
        Con_DebugLog ("%s: lump size %d is not a multiple of %d in %s\n", Function, l->filelen, Size, loadmodel->name);
        exit(1);
    }
}


/*
=================
Mod_ChkBounds
=================
*/
static void Mod_ChkBounds (char *Function, char *Object, int SItem, int Items, int Limit)
{
    int ErrItem;

    if (!Mod_WithinBounds (SItem, Items, Limit, &ErrItem))
    {
        // Should be an error but ...
        // we took the silent approach :)
        Con_DebugLog ("%s: %s is out of bounds (%d, max = %d) in %s\n", Function, Object, ErrItem, Limit, loadmodel->name);
    }
}

/*
=================
Mod_ChkFileSize
=================
*/
static void Mod_ChkFileSize (char *Function, char *Object, int SItem, int Items)
{
    int ErrItem;

    if (!Mod_WithinBounds (SItem, Items, mod_filesize, &ErrItem))
    {
        // Should be an error but ...
        // we took the silent approach :)
        Con_DebugLog ("%s: %s is outside file (%d, max = %d) in %s\n", Function, Object, ErrItem, mod_filesize, loadmodel->name);
    }
}

/*
=================
Mod_ChkType
=================
*/
static void Mod_ChkType (char *Function, char *Object, char *ObjNum, int Type, int Type1, int Type2)
{
    if (Type != Type1 && Type != Type2)
    {
        // Should be an error but ...
        // we took the silent approach :)
        Con_DebugLog ("invalid %s %d in %s in %s\n", Object, Type, ObjNum, loadmodel->name);
    }
}

/*
=================
Mod_LoadTextures
=================
*/
void Mod_LoadTextures (lump_t *l)
{
    int				i, j, pixels, num, max, altmax;
    miptex_t		*mt;
    texture_t		*tx, *tx2;
    texture_t		*anims[10];
    texture_t		*altanims[10];
    dmiptexlump_t	*m;

    if (!l->filelen)
    {
        loadmodel->textures = NULL;
        return;
    }

    // Check bounds for nummiptex var
    Mod_ChkBounds ("Mod_LoadTextures", "nummiptex", 0, sizeof(int), l->filelen);

    m = (dmiptexlump_t *)(mod_base + l->fileofs);
    m->nummiptex = LittleLong (m->nummiptex);

    // Check bounds for dataofs array
    if (m->nummiptex > 0)
    {
        Mod_ChkBounds ("Mod_LoadTextures", "miptex lump", sizeof(int), m->nummiptex * sizeof(int), l->filelen);
    }
    loadmodel->numtextures = m->nummiptex;
    loadmodel->textures = Heap_ZAlloc (m->nummiptex * sizeof(*loadmodel->textures));

    for (i=0 ; i<m->nummiptex ; i++)
    {
        m->dataofs[i] = LittleLong(m->dataofs[i]);

        if (m->dataofs[i] == -1) continue;

        mt = (miptex_t *)((byte *)m + m->dataofs[i]);

        // Check bounds for miptex entry
        Mod_ChkBounds ("Mod_LoadTextures", "miptex entry", (byte *)mt - (mod_base + l->fileofs), sizeof(miptex_t), l->filelen);

        mt->width = LittleLong (mt->width);
        mt->height = LittleLong (mt->height);

        if (mt->width <= 0 || mt->height <= 0 || (mt->width & 15) || (mt->height & 15))
        {
            Sys_Error ("Mod_LoadTextures: texture '%s' is not 16 aligned (%dx%d) in %s", mt->name, mt->width, mt->height, loadmodel->name);
        }
        pixels = mt->width*mt->height/64*85;
        tx = Heap_ZAlloc (sizeof(texture_t) + pixels);
        loadmodel->textures[i] = tx;

        // store the texture number for volume carving
        tx->num = i;

        memcpy (tx->name, mt->name, sizeof (tx->name));

        tx->width = mt->width;
        tx->height = mt->height;

        // HACK HACK HACK
        // Note - CRC doesn't guarantee uniqueness, but it'll do for a hack
        if (!strcmp (mt->name, "shot1sid") && mt->width == 32 && mt->height == 32 && CRC_Block ((byte *)(mt + 1), mt->width * mt->height) == 65393)
        {
            // This texture in b_shell1.bsp has some of the first 32 pixels painted white.
            // They are invisible in software, but look really ugly in GL. So we just copy
            // 32 pixels from the bottom to make it look nice.
            memcpy (mt + 1, (byte *) (mt + 1) + 32 * 31, 32);
        }

        // the pixels immediately follow the structures
        memcpy (tx+1, mt+1, pixels);

        // set the data pointer for loading
        tx->data = (byte *) (tx + 1);

        if (!strncmp (mt->name, "sky", 3))
        {
            R_InitSky (tx);
        }
        else if (mt->name[0] == '*')
        {
            tx->gl_texture = GL_LoadQuakeTexture (NULL, tx->name, tx->data, tx->width, tx->height, (TEX_MIPMAP | TEX_BSP));
        }
        else
        {
            tx->gl_texture = GL_LoadQuakeTexture (NULL, tx->name, tx->data, tx->width, tx->height, (TEX_MIPMAP | TEX_BSP | TEX_MAKENORMAL));
        }
    }

    // sequence the animations
    MissFrameNo = 0;
    memset (MissFrameName, 0, sizeof(MissFrameName));

    for (i=0 ; i<m->nummiptex ; i++)
    {
        tx = loadmodel->textures[i];
        if (!tx || tx->name[0] != '+') continue;
        if (tx->anim_next) continue;	// allready sequenced

        // find the number of frames in the animation
        memset (anims, 0, sizeof(anims));
        memset (altanims, 0, sizeof(altanims));

        max = tx->name[1];
        altmax = 0;

        if (max >= 'a' && max <= 'z')
        {
            max -= 'a' - 'A';
        }

        if (max >= '0' && max <= '9')
        {
            max -= '0';
            altmax = 0;
            anims[max] = tx;
            max++;
        }
        else if (max >= 'A' && max <= 'J')
        {
            altmax = max - 'A';
            max = 0;
            altanims[altmax] = tx;
            altmax++;
        }
        else
        {
            Sys_Error ("Bad animating texture %s", tx->name);
        }

        for (j=i+1 ; j<m->nummiptex ; j++)
        {
            tx2 = loadmodel->textures[j];
            if (!tx2 || tx2->name[0] != '+') continue;
            if (strcmp (tx2->name+2, tx->name+2)) continue;

            num = tx2->name[1];

            if (num >= 'a' && num <= 'z')
            {
                num -= 'a' - 'A';
            }

            if (num >= '0' && num <= '9')
            {
                num -= '0';
                anims[num] = tx2;

                if (num+1 > max)
                {
                    max = num + 1;
                }
            }
            else if (num >= 'A' && num <= 'J')
            {
                num = num - 'A';
                altanims[num] = tx2;

                if (num + 1 > altmax)
                {
                    altmax = num+1;
                }
            }
            else
            {
                Sys_Error ("Bad animating texture %s", tx->name);
            }
        }

        // Check if all frames are present if not disable animation sequence
        if (!Mod_ChkAllFrames (tx, anims, max, false) || !Mod_ChkAllFrames (tx, altanims, altmax, true)) continue;

#define	ANIM_CYCLE	2
        // link them all together
        for (j=0 ; j<max ; j++)
        {
            tx2 = anims[j];
            tx2->anim_total = max * ANIM_CYCLE;
            tx2->anim_min = j * ANIM_CYCLE;
            tx2->anim_max = (j+1) * ANIM_CYCLE;
            tx2->anim_next = anims[ (j+1)%max ];
            if (altmax)	tx2->alternate_anims = altanims[0];
        }

        for (j=0 ; j<altmax ; j++)
        {
            tx2 = altanims[j];
            tx2->anim_total = altmax * ANIM_CYCLE;
            tx2->anim_min = j * ANIM_CYCLE;
            tx2->anim_max = (j+1) * ANIM_CYCLE;
            tx2->anim_next = altanims[ (j+1)%altmax ];
            if (max) tx2->alternate_anims = anims[0];
        }
    }
}

/*
=================
Mod_LoadBSPLighting

Contains some evil hackery to rescale the colour balance of lightmaps so that they preserve the same
intensity as the original BSP lighting.  This is needed cos various versions of MHColour didn't quite
do the job properly, and it's too late to go back now...
=================
*/
void Mod_LoadBSPLighting (lump_t *l)
{
    int				i;				// loop control
    int				j;
    byte			*light_base;	// base offset of light data
    filehandle_t	LITFile;
    int				LITSize;
    char			LITFileName[1024];
    litheader_t		LITHeader;
    byte			*light_buffer;
    int				bspintens;
    int				litintens;
    int				templ;
    int				hit;
    int				ntscscales[3] = {30, 59, 11};

    if (!l->filelen)
    {
        loadmodel->lightdata = NULL;
        return;
    }

    // load from the BSP and expand to 3 component
    loadmodel->lightdata = Heap_ZAlloc (l->filelen * 3);
    light_base = mod_base + l->fileofs;

    for (i = 0; i < l->filelen; i++)
    {
        // unnecessary brackets improve legibility and clarity
        loadmodel->lightdata[(i * 3) + 0] = light_base[i];
        loadmodel->lightdata[(i * 3) + 1] = light_base[i];
        loadmodel->lightdata[(i * 3) + 2] = light_base[i];
    }

    // build the name of the lit file
    strlcpy_s(LITFileName, loadmodel->name);

    // we can't just search for '.' cos there may be one in a folder name!!!
    for (i = (int) strlen (LITFileName) - 1; i; i--)
    {
        if (LITFileName[i] == '.')
        {
            LITFileName[i + 1] = 'l';
            LITFileName[i + 2] = 'i';
            LITFileName[i + 3] = 't';

            // done
            break;
        }
    }

    // try to open it
    LITSize = COM_FOpenFile (LITFileName, &LITFile, FS_READ);

    // no lit file
    if (!LITFile) return;

    // not from correct dir
    if (LITSize - 8 != l->filelen * 3)
    {
        // invalid lit file for this map
        COM_FCloseFile (LITFile);
        return;
    }

    // read it in
    COM_Read (&LITHeader, sizeof (LITHeader), LITFile);

    // validate the header
    if (!(LITHeader.version == 1 && LITHeader.ident[0] == 'Q' && LITHeader.ident[1] == 'L' && LITHeader.ident[2] == 'I' && LITHeader.ident[3] == 'T'))
    {
        // bad lit file
        COM_FCloseFile (LITFile);
        return;
    }

    // it's valid
    light_buffer = (byte *) Heap_MemAlloc (l->filelen * 3);

    // read it in
    COM_Read (light_buffer, l->filelen * 3, LITFile);

    // now compare with the light from the BSP - incorrect lightmap sizes in
    // mhcolour can cause some weird effects in some areas - the nailgun room
    // in e1m1 is a perfect example.  this is a hack - the real solution
    // would be to rewrite mhcolour to fix this.  too late for that though...
    for (i = 0; i < l->filelen; i++)
    {
        bspintens = 0;
        litintens = 0;
        hit = 0;

        for (j = 0; j < 3; j++)
        {
            // skip LIT loading on these conditions
            if (light_buffer[(i * 3) + j] == loadmodel->lightdata[(i * 3) + j]) continue;
            if (light_buffer[(i * 3) + j] == 0) continue;
            if (loadmodel->lightdata[(i * 3) + j] == 0) continue;
            if (light_buffer[(i * 3) + j] == 255) continue;
            if (loadmodel->lightdata[(i * 3) + j] == 255) continue;

            // load it in, and store the scales
            bspintens += (int) loadmodel->lightdata[(i * 3) + j] * ntscscales[j];
            litintens += (int) light_buffer[(i * 3) + j] * ntscscales[j];
            hit++;

            loadmodel->lightdata[(i * 3) + j] = light_buffer[(i * 3) + j];
        }

        // we need 3 hits for the next stage
        if (hit != 3) continue;

        // other continue conditions
        if (bspintens == litintens) continue;
        if (!bspintens) continue;
        if (!litintens) continue;

        // rescale the lit file to match the original in the bsp
        for (j = 0; j < 3; j++)
        {
            // scale
            templ = loadmodel->lightdata[(i * 3) + j];
            templ *= bspintens;
            templ /= litintens;

            // bound
            if (templ > 255)
            {
                templ = 255;
            }

            if (templ < 0)
            {
                templ = 0;
            }

            // store
            loadmodel->lightdata[(i * 3) + j] = (byte) templ;
        }
    }
    Heap_MemAllocFree (light_buffer);
    COM_FCloseFile (LITFile);
}

/*
=================
Mod_LoadLighting
=================
*/
void Mod_LoadLighting (lump_t *l)
{
    int				i;
    unsigned short	*out;

    // call to Mod_LoadBSPLighting (which was the previous lighting loader) as a seperate func
    // so that we can set stainmaps correctly irrespective of the return point from that one.
    Mod_LoadBSPLighting (l);

    // set stainmaps to NULL if there's no lightdata.  this is unfortunate, but we really do need
    // the light offsets in the BSP.
    if (!l->filelen)
    {
        loadmodel->staindata = NULL;
        return;
    }
    // now set up staindata
    loadmodel->staindata = Heap_ZAlloc (l->filelen * 3 * sizeof (short));
    out = loadmodel->staindata;

    // clear to nothing
    for (i = 0; i < l->filelen; i++)
    {
        *out++ = 0;
        *out++ = 0;
        *out++ = 0;
    }
}

/*
=================
Mod_LoadEntities
=================
*/
void Mod_LoadEntities (lump_t *l)
{
    if (!l->filelen)
    {
        loadmodel->entities = NULL;
        loadmodel->numentities = 0;
        return;
    }

    // add 1 to allow for NULL termination
    loadmodel->entities = Heap_ZAlloc (l->filelen + 1);

    // clear the string
    memset (loadmodel->entities, 0, l->filelen + 1);

    // copy it in
    memcpy (loadmodel->entities, mod_base + l->fileofs, l->filelen);
    loadmodel->numentities = l->filelen;
}

/*
=================
Mod_LoadVertexes
=================
*/
void Mod_LoadVertexes (lump_t *l)
{
    dvertex_t	*in;
    mvertex_t	*out;
    int			i, j, count;

    in = (void *)(mod_base + l->fileofs);
    Mod_ChkLumpSize ("Mod_LoadVertexes", l, sizeof(*in));
    count = l->filelen / sizeof(*in);

    if (count > MAX_MAP_VERTS)
    {
        Con_SafePrintf ("\x02Mod_LoadVertexes: ");
        Con_Printf ("excessive vertexes (%d, normal max = %d) in %s\n", count, MAX_MAP_VERTS, loadmodel->name);
    }
    out = Heap_ZAlloc (count * sizeof(*out));

    loadmodel->vertexes = out;
    loadmodel->numvertexes = count;

    for ( i=0 ; i<count ; i++, in++, out++)
    {
        for (j = 0; j < 3; ++j)
        {
            out->position[j] = LittleFloat (in->point[j]);

            // Check bounds
            if (fabs(out->position[j]) > 65536)
            {
                Con_SafePrintf ("\x02Mod_LoadVertexes: ");
                Con_SafePrintf ("vertex %d,%d way out of bounds (%.0f, max = %d) in %s\n", i, j, out->position[j], 4096, loadmodel->name);
                out->position[j] = 0; // OK?
            }
        }
    }
}

/*
=================
Mod_LoadSubmodels
=================
*/
void Mod_LoadSubmodels (lump_t *l)
{
    dmodel_t	*in;
    dmodel_t	*out;
    int			i, j, count, firstface;

    in = (void *)(mod_base + l->fileofs);
    Mod_ChkLumpSize ("Mod_LoadSubmodels", l, sizeof(*in));
    count = l->filelen / sizeof(*in);

    if (count > MAX_MODELS) // Why not MAX_MAP_MODELS?
    {
        Sys_Error ("Mod_LoadSubmodels: too many models (%d, max = %d) in %s", count, MAX_MODELS, loadmodel->name);
    }

    if (count > MAX_MAP_MODELS) // Old limit
    {
        Con_SafePrintf ("\x02Mod_LoadSubmodels: ");
        Con_Printf ("excessive models (%d, normal max = %d) in %s\n", count, MAX_MAP_MODELS, loadmodel->name);
    }
    out = Heap_ZAlloc ( count*sizeof(*out));

    loadmodel->submodels = out;
    loadmodel->numsubmodels = count;

    for ( i=0 ; i<count ; i++, in++, out++)
    {
        for (j=0 ; j<3 ; j++)
        {
            // spread the mins / maxs by a pixel
            out->mins[j] = LittleFloat (in->mins[j]) - 1;
            out->maxs[j] = LittleFloat (in->maxs[j]) + 1;
            out->origin[j] = LittleFloat (in->origin[j]);
        }

        for (j=0 ; j<MAX_MAP_HULLS ; j++)
        {
            out->headnode[j] = LittleLong (in->headnode[j]);
        }
        out->visleafs = LittleLong (in->visleafs);
        firstface = LittleLong (in->firstface);
        out->numfaces = LittleLong (in->numfaces);

        // Check bounds
        Mod_ChkBounds ("Mod_LoadSubmodels", "face", firstface, out->numfaces, loadmodel->numsurfaces);

        out->firstface = firstface;
    }

    // Check world visleafs
    out = loadmodel->submodels;

    if (out->visleafs > MAX_MAP_LEAFS)
    {
        Sys_Error ("Mod_LoadSubmodels: too many visleafs (%d, max = %d) in %s", out->visleafs, MAX_MAP_LEAFS, loadmodel->name);
    }
}

/*
=================
Mod_LoadEdges
=================
*/
void Mod_LoadEdges (lump_t *l)
{
    dedge_t *in;
    medge_t *out;
    int 	i, j, count;

    in = (void *)(mod_base + l->fileofs);
    Mod_ChkLumpSize ("Mod_LoadEdges", l, sizeof(*in));
    count = l->filelen / sizeof(*in);
    out = Heap_ZAlloc ( (count + 1) * sizeof(*out));

    loadmodel->edges = out;
    loadmodel->numedges = count;

    for ( i=0 ; i<count ; i++, in++, out++)
    {
        for (j = 0; j < 2; ++j)
        {
            out->v[j] = (unsigned short)LittleShort (in->v[j]);

            if (i == 0 && j == 0) continue; // Why is sometimes this edge invalid?

            // Check bounds
            if (out->v[j] >= loadmodel->numvertexes)
            {
                Con_SafePrintf ("\x02Mod_LoadEdges: ");
                Con_SafePrintf ("vertex %d in edge %d out of bounds (%d, max = %d) in %s\n", j, i, out->v[j], loadmodel->numvertexes, loadmodel->name);
                out->v[j] = 0; // OK?
            }
        }
    }
}

/*
=================
Mod_LoadTexinfo
=================
*/
void Mod_LoadTexinfo (lump_t *l)
{
    texinfo_t	*in;
    mtexinfo_t	*out;
    int			i, j, count;
    int			miptex;
    float		len1, len2;

    in = (void *) (mod_base + l->fileofs);
    Mod_ChkLumpSize ("Mod_LoadTexinfo", l, sizeof(*in));
    count = l->filelen / sizeof (*in);

    if (count > MAX_MAP_TEXINFO)
    {
        Con_SafePrintf ("\x02Mod_LoadTexinfo: ");
        Con_Printf ("excessive texinfo (%d, normal max = %d) in %s\n", count, MAX_MAP_TEXINFO, loadmodel->name);
    }
    out = Heap_ZAlloc ( count*sizeof(*out));

    loadmodel->texinfo = out;
    loadmodel->numtexinfo = count;

    for (i = 0; i < count; i++, in++, out++)
    {
        for (j = 0; j < 8; j++) out->vecs[0][j] = LittleFloat (in->vecs[0][j]);

        len1 = Length (out->vecs[0]);
        len2 = Length (out->vecs[1]);
        len1 = (len1 + len2) / 2;

        if (len1 < 0.32)
        {
            out->mipadjust = 4;
        }
        else if (len1 < 0.49)
        {
            out->mipadjust = 3;
        }
        else if (len1 < 0.99)
        {
            out->mipadjust = 2;
        }
        else
        {
            out->mipadjust = 1;
        }
        miptex = LittleLong (in->miptex);

        out->flags = LittleLong (in->flags);

        if (!loadmodel->textures)
        {
            // checkerboard texture
            out->texture = r_notexture_mip;
            out->flags = 0;
        }
        else
        {
            // MH - this was a Sys_Error
            if (miptex >= loadmodel->numtextures)
            {
                goto SetNoTexture;
            }
            out->texture = loadmodel->textures[miptex];

            if (!out->texture)
            {
SetNoTexture:
                ;
                // texture not found
                out->texture = r_notexture_mip;
                out->flags = 0;
            }
        }
    }
}

/*
================
CalcSurfaceExtents

Fills in s->texturemins[] and s->extents[]
================
*/
void CalcSurfaceExtents (msurface_t *s)
{
    float		mins[2], maxs[2], val;
    int			i, j, e;
    mvertex_t	*v;
    mtexinfo_t	*tex;
    int			bmins[2], bmaxs[2];

    mins[0] = mins[1] = 999999;
    maxs[0] = maxs[1] = -99999;

    tex = s->texinfo;

    for (i=0 ; i<s->numedges ; i++)
    {
        e = loadmodel->surfedges[s->firstedge+i];

        if (e >= 0)
        {
            v = &loadmodel->vertexes[loadmodel->edges[e].v[0]];
        }
        else
        {
            v = &loadmodel->vertexes[loadmodel->edges[-e].v[1]];
        }

        for (j=0 ; j<2 ; j++)
        {
            val = v->position[0] * tex->vecs[j][0] +
                  v->position[1] * tex->vecs[j][1] +
                  v->position[2] * tex->vecs[j][2] +
                  tex->vecs[j][3];

            if (val < mins[j])
            {
                mins[j] = val;
            }

            if (val > maxs[j])
            {
                maxs[j] = val;
            }
        }
    }

    for (i=0 ; i<2 ; i++)
    {
        bmins[i] = floor(mins[i]/16);
        bmaxs[i] = ceil(maxs[i]/16);

        s->texturemins[i] = bmins[i] * 16;
        s->extents[i] = (bmaxs[i] - bmins[i]) * 16;

        if (!(tex->flags & TEX_SPECIAL) && s->extents[i] > MAX_SURF_EXTENTS)
        {
            Con_SafePrintf ("\002CalcSurfaceExtents: ");
            Con_SafePrintf ("excessive surface extents (%d, normal max = %d), texture %s in %s\n", s->extents[i], MAX_SURF_EXTENTS, tex->texture->name, loadmodel->name);
            s->extents[i] = MAX_SURF_EXTENTS; // Kludge, but seems to work
        }
    }
}


/*
=================
Mod_LoadFaces
=================
*/
void Mod_LoadFaces (lump_t *l)
{
    dface_t		*fa;
    msurface_t 	*surf;
    int			i, count, surfnum, texinfo, firstsurfedge;
    int			planenum, side;

    fa = (void *) (mod_base + l->fileofs);
    Mod_ChkLumpSize ("Mod_LoadFaces", l, sizeof(*fa));
    count = l->filelen / sizeof (*fa);

    if (count > MAX_MAP_FACES)
    {
        Con_SafePrintf ("\x02Mod_LoadFaces: ");
        Con_Printf ("excessive faces (%d, normal max = %d) in %s\n", count, MAX_MAP_FACES, loadmodel->name);
    }
    surf = Heap_ZAlloc (count * sizeof (*surf));

    loadmodel->surfaces = surf;
    loadmodel->numsurfaces = count;

    for ( surfnum=0 ; surfnum<count ; surfnum++, fa++, surf++)
    {
        firstsurfedge = LittleLong (fa->firstedge);
        surf->numedges = LittleShort(fa->numedges);

        // Check bounds
        Mod_ChkBounds ("MOD_LoadFaces", "surfedge", firstsurfedge, surf->numedges, loadmodel->numsurfedges);

        surf->firstedge = firstsurfedge;
        surf->flags = 0;

        planenum = LittleShort(fa->planenum);

        Mod_ChkBounds ("MOD_LoadFaces", "plane", planenum, 1, loadmodel->numplanes);

        side = LittleShort(fa->side);

        if (side)
        {
            surf->flags |= SURF_PLANEBACK;
        }
        surf->plane = loadmodel->planes + planenum;
        texinfo = LittleShort (fa->texinfo);

        Mod_ChkBounds ("MOD_LoadFaces", "texinfo", texinfo, 1, loadmodel->numtexinfo);

        surf->texinfo = loadmodel->texinfo + texinfo;

        CalcSurfaceExtents (surf);

        // set up light styles
        for (i = 0; i < MAXLIGHTMAPS; i++)
        {
            surf->styles[i] = fa->styles[i];
        }
        i = LittleLong (fa->lightofs);

        // replicated surf->samples for surf->stainsamples
        // also fixed up where loadmodel->lightdata may be NULL but i may have a value from
        // either a crappy BSP compiler or something else going wrong somewhere
        if (i == -1 || !loadmodel->lightdata)
        {
            surf->samples = NULL;
            surf->stainsamples = NULL;
        }
        else
        {
            surf->samples = loadmodel->lightdata + (i * 3);
            surf->stainsamples = loadmodel->staindata + (i * 3);
        }

        // store the surface number
        surf->surfnum = surfnum;

        // set an initial valid lightmap pointer.
        // as stainmaps will be stored in the same struct, this will suffice for both
        surf->lightmap = NULL;
        surf->volume = -1;
        surf->volumetexture = NULL;

        // precalculate the surface warp indexes
        // this is valid for liquid and for normal surfs (needed for caustics)
        // we really should be setting a plane coefficient for either the texture matrix or some
        // texgen stuff based on surf->truenormal here, but bleagh.  maybe someday...
        switch (surf->plane->type)
        {
        case PLANE_Y:
            // vertical in line with the world grid
            surf->warpindex[0] = 0;
            surf->warpindex[1] = 2;
            break;

        case PLANE_X:
            // vertical in line with the world grid
            surf->warpindex[0] = 2;
            surf->warpindex[1] = 1;
            break;

        case PLANE_ANYX:
        case PLANE_ANYY:
            // vertical at an angle to the world grid
            surf->warpindex[0] = 2;
            surf->warpindex[1] = 0;
            break;

        case PLANE_Z:
            // flat
        default:
            // same as flat
            surf->warpindex[0] = 0;
            surf->warpindex[1] = 1;
            break;
        }

        // catch e2m3 crapness
        if (surf->texinfo->texture->name[0] == 0)
        {
            surf->texinfo->texture->gl_texture = &r_black;
        }

        // set the drawing flags flag
        if (!strncasecmp(surf->texinfo->texture->name, "sky", 3))	// sky
        {
            surf->flags |= (SURF_DRAWSKY | SURF_DRAWTILED);
            continue;
        }

        if (surf->texinfo->texture->gl_texture->flags & TEX_LIQUID)
        {
            surf->flags |= (SURF_DRAWTURB | SURF_DRAWTILED);

            // set the contents type here
            if (surf->texinfo->texture->name[1] == 'l' &&
                surf->texinfo->texture->name[2] == 'a' &&
                surf->texinfo->texture->name[3] == 'v' &&
                surf->texinfo->texture->name[4] == 'a')
                surf->flags |= SURF_DRAWLAVA;
            else if (surf->texinfo->texture->name[1] == 't' &&
                surf->texinfo->texture->name[2] == 'e' &&
                surf->texinfo->texture->name[3] == 'l' &&
                surf->texinfo->texture->name[4] == 'e')
                surf->flags |= SURF_DRAWTELE;
            else if (surf->texinfo->texture->name[1] == 's' &&
                surf->texinfo->texture->name[2] == 'l' &&
                surf->texinfo->texture->name[3] == 'i' &&
                surf->texinfo->texture->name[4] == 'p')
                surf->flags |= SURF_DRAWTELE;
            else if (surf->texinfo->texture->name[1] == 's' &&
                surf->texinfo->texture->name[2] == 'l' &&
                surf->texinfo->texture->name[3] == 'i' &&
                surf->texinfo->texture->name[4] == 'm' &&
                surf->texinfo->texture->name[5] == 'e')
                surf->flags |= SURF_DRAWSLIME;
            else surf->flags |= SURF_DRAWWATER;

            for (i = 0; i < 2; i++)
            {
                surf->extents[i] = 16384;
                surf->texturemins[i] = -8192;
            }
            continue;
        }
    }
}

/*
=================
Mod_SetParent
=================
*/
void Mod_SetParent (mnode_t *node, mnode_t *parent)
{
    node->parent = parent;
    if (node->contents < 0)	return;
    Mod_SetParent (node->children[0], node);
    Mod_SetParent (node->children[1], node);
}

/*
=================
Mod_LoadNodes
=================
*/
void Mod_LoadNodes (lump_t *l)
{
    int			i, j, count, p, firstface;
    dnode_t		*in;
    mnode_t 	*out;

    in = (void *)(mod_base + l->fileofs);
    Mod_ChkLumpSize ("Mod_LoadNodes", l, sizeof(*in));
    count = l->filelen / sizeof(*in);

    // Check nodes
    if (count > MAX_MAP_NODES)
    {
        Con_SafePrintf ("\002Mod_LoadNodes: ");
        Con_Printf ("excessive nodes (%d, normal max = %d) in %s\n", count, MAX_MAP_NODES, loadmodel->name);
    }
    out = Heap_ZAlloc (count * sizeof(*out));

    loadmodel->nodes = out;
    loadmodel->numnodes = count;

    for ( i=0 ; i<count ; i++, in++, out++)
    {
        for (j=0 ; j<3 ; j++)
        {
            out->minmaxs[j] = LittleShort (in->mins[j]);
            out->minmaxs[3+j] = LittleShort (in->maxs[j]);
        }
        p = LittleLong(in->planenum);

        Mod_ChkBounds ("MOD_LoadNodes", "plane", p, 1, loadmodel->numplanes);

        out->plane = loadmodel->planes + p;
        firstface = LittleShort (in->firstface);
        out->numsurfaces = LittleShort (in->numfaces);

        // Check bounds
        Mod_ChkBounds ("MOD_LoadNodes", "face", firstface, out->numsurfaces, loadmodel->numsurfaces);

        out->firstsurface = firstface;

        for (j=0 ; j<2 ; j++)
        {
            p = LittleShort (in->children[j]);

            if (p < -loadmodel->numleafs)
            {
                p += MAX_MAP_PORTALS; // Gross hack to connect as much as possible of the crippled bsp tree
            }

            if (p >= 0)
            {
                if (p >= loadmodel->numnodes)
                {
                    p = loadmodel->numnodes - 1;
                }
                out->children[j] = loadmodel->nodes + p;
            }
            else
            {
                if ((-1 - p) >= loadmodel->numleafs)
                {
                    p = -loadmodel->numleafs;
                }
                out->children[j] = (mnode_t *)(loadmodel->leafs + (-1 - p));
            }
        }
    }
    Mod_SetParent (loadmodel->nodes, NULL);	// sets nodes and leafs
}

/*
=================
Mod_LoadLeafs
=================
*/
void Mod_LoadLeafs (lump_t *l, lump_t *v)
{
    dleaf_t		*in;
    mleaf_t		*out;
    byte		*vis;
    int			i, j, count, p, firstmarksurface;

    in = (void *)(mod_base + l->fileofs);
    Mod_ChkLumpSize ("Mod_LoadLeafs", l, sizeof(*in));
    count = l->filelen / sizeof(*in);

    // Check leafs
    if (count > MAX_MAP_LEAFS)
    {
        Sys_Error ("MOD_LoadLeafs: too many leafs (%d, max = %d) in %s", count, 32767, loadmodel->name);
    }
    out = Heap_ZAlloc (count * sizeof(*out));

    loadmodel->leafs = out;
    loadmodel->numleafs = count;

    // Mod_LoadVisibility stuff
    // we know how many leafs there are in the map, so we can set up the visdata now.
    if (v->filelen)
    {
        // add 7 to prevent round down errors with int division by 8.
        loadmodel->vissize = (loadmodel->numleafs + 7) / 8;

        // load it in
        loadmodel->visdata = Heap_ZAlloc (loadmodel->numleafs * loadmodel->vissize);

        // get lump visibility count we need this further down.
        loadmodel->numvisdata = v->filelen;

        // set to all solid initially
        memset (loadmodel->visdata, 0xff, loadmodel->numleafs * loadmodel->vissize);
    }
    else
    {
        // no visible data
        loadmodel->vissize = 0;
        loadmodel->visdata = NULL;
    }

    for (i = 0 ; i < count ; i++, in++, out++)
    {
        for (j=0 ; j<3 ; j++)
        {
            out->mins[j] = LittleShort (in->mins[j]);
            out->maxs[j] = LittleShort (in->maxs[j]);
        }
        p = LittleLong(in->contents);
        out->contents = p;

        firstmarksurface = (unsigned short)LittleShort (in->firstmarksurface);
        out->nummarksurfaces = (unsigned short)LittleShort (in->nummarksurfaces);

        // Check bounds
        Mod_ChkBounds ("MOD_LoadLeafs", "marksurfaces", firstmarksurface, out->nummarksurfaces, loadmodel->nummarksurfaces);

        out->firstmarksurface = loadmodel->marksurfaces + firstmarksurface;

        p = LittleLong(in->visofs);

        if (p != -1)
        {
            if (!loadmodel->visdata)
            {
                p = -1;
            }
            else if (!Mod_WithinBounds (p, 1, loadmodel->numvisdata, NULL))
            {
                if (i > 0) // Don't warn about leaf 0, sometimes bad (?)
                {
                    Con_SafePrintf ("\002MOD_LoadLeafs: ");
                    Con_SafePrintf ("visofs in leaf %d out of bounds (%d, max = %d) in %s\n", i, p, loadmodel->numvisdata, loadmodel->name);
                }
                p = -1;
            }
        }

        // removed leaf checks here
        // we dont use the auto water trans code !!! this is for volume carving only.
        if (p != -1)
        {
            // paranoid about brackets!!!
            out->decompressed_vis = loadmodel->visdata + (i * loadmodel->vissize);

            // decompress visdata
            vis = Mod_DecompressVis (mod_base + v->fileofs + p, loadmodel);

            // copy it in
            memcpy (out->decompressed_vis, vis, loadmodel->vissize);
        }
        else
        {
            out->decompressed_vis = NULL;
        }
        out->efrags = NULL;
        out->volume = -1;
        out->volumetexture = NULL;

        for (j=0 ; j<4 ; j++)
        {
            out->ambient_sound_level[j] = in->ambient_level[j];
        }

        // gl underwater warp
        if (CONTENTMASK_LIQUID(out))
        {
            for (j=0 ; j<out->nummarksurfaces ; j++)
            {
                out->firstmarksurface[j]->flags |= SURF_UNDERWATER;
            }
        }
    }
}

/*
=================
Mod_LoadClipnodes
=================
*/
void Mod_LoadClipnodes (lump_t *l)
{
    dclipnode_t *in, *out;
    int			i, count;
    hull_t		*hull;

    in = (void *)(mod_base + l->fileofs);
    Mod_ChkLumpSize ("Mod_LoadClipnodes", l, sizeof(*in));
    count = l->filelen / sizeof(*in);

    if (count > MAX_MAP_CLIPNODES) // Normal limit
    {
        Con_SafePrintf ("\x02Mod_LoadClipnodes: ");
        Con_Printf ("excessive clipnodes (%d, normal max = %d) in %s\n", count, MAX_MAP_CLIPNODES, loadmodel->name);
    }
    out = Heap_ZAlloc (count * sizeof(*out));

    loadmodel->clipnodes = out;
    loadmodel->numclipnodes = count;

    hull = &loadmodel->hulls[1];
    hull->clipnodes = out;
    hull->firstclipnode = 0;
    hull->lastclipnode = count-1;
    hull->planes = loadmodel->planes;
    hull->clip_mins[0] = -16;
    hull->clip_mins[1] = -16;
    hull->clip_mins[2] = -24;
    hull->clip_maxs[0] = 16;
    hull->clip_maxs[1] = 16;
    hull->clip_maxs[2] = 32;

    hull = &loadmodel->hulls[2];
    hull->clipnodes = out;
    hull->firstclipnode = 0;
    hull->lastclipnode = count-1;
    hull->planes = loadmodel->planes;
    hull->clip_mins[0] = -32;
    hull->clip_mins[1] = -32;
    hull->clip_mins[2] = -24;
    hull->clip_maxs[0] = 32;
    hull->clip_maxs[1] = 32;
    hull->clip_maxs[2] = 64;

    for (i=0 ; i<count ; i++, out++, in++)
    {
        out->planenum = LittleLong(in->planenum);

        // Check bounds
        if (out->planenum < 0 || out->planenum >= loadmodel->numplanes)
        {
            Con_SafePrintf ("\x02Mod_LoadClipnodes: ");
            Con_SafePrintf ("plane in clipnode %d out of bounds (%d, max = %d) in %s\n", i, out->planenum, loadmodel->numplanes, loadmodel->name);
            out->planenum = 0; // Kludge, but seems to work
        }
        out->children[0] = LittleShort(in->children[0]);
        out->children[1] = LittleShort(in->children[1]);
    }
}

/*
=================
Mod_MakeHull0

Deplicate the drawing hull structure as a clipping hull
=================
*/
void Mod_MakeHull0 (void)
{
    mnode_t		*in, *child;
    dclipnode_t *out;
    int			i, j, count;
    hull_t		*hull;

    hull = &loadmodel->hulls[0];

    in = loadmodel->nodes;
    count = loadmodel->numnodes;
    out = Heap_ZAlloc ( count*sizeof(*out));

    hull->clipnodes = out;
    hull->firstclipnode = 0;
    hull->lastclipnode = count-1;
    hull->planes = loadmodel->planes;

    for (i=0 ; i<count ; i++, out++, in++)
    {
        out->planenum = in->plane - loadmodel->planes;

        for (j=0 ; j<2 ; j++)
        {
            child = in->children[j];

            if (child->contents < 0)
            {
                out->children[j] = child->contents;
            }
            else
            {
                out->children[j] = child - loadmodel->nodes;
            }
        }
    }
}

/*
=================
Mod_LoadMarksurfaces
=================
*/
void Mod_LoadMarksurfaces (lump_t *l)
{
    int			i, j, count;
    short		*in;
    msurface_t	**out;

    in = (void *)(mod_base + l->fileofs);
    Mod_ChkLumpSize ("Mod_LoadMarksurfaces", l, sizeof(*in));
    count = l->filelen / sizeof(*in);

    if (count > MAX_MAP_MARKSURFACES) // Normal limit
    {
        Con_SafePrintf ("\x02Mod_LoadMarksurfaces: ");
        Con_Printf ("excessive marksurfaces (%d, normal max = %d) in %s\n", count, MAX_MAP_MARKSURFACES, loadmodel->name);
    }
    out = Heap_ZAlloc (count * sizeof(*out));

    loadmodel->marksurfaces = out;
    loadmodel->nummarksurfaces = count;

    for (i=0 ; i<count ; i++)
    {
        j = LittleShort(in[i]);
        Mod_ChkBounds ("Mod_LoadMarksurfaces", "face", j, 1, loadmodel->numsurfaces);
        out[i] = loadmodel->surfaces + j;
    }
}

/*
=================
Mod_LoadSurfedges
=================
*/
void Mod_LoadSurfedges (lump_t *l)
{
    int		i, count;
    int		*in, *out;

    in = (void *)(mod_base + l->fileofs);
    Mod_ChkLumpSize ("Mod_LoadSurfedges", l, sizeof(*in));
    count = l->filelen / sizeof(*in);
    out = Heap_ZAlloc ( count*sizeof(*out));

    loadmodel->surfedges = out;
    loadmodel->numsurfedges = count;

    for (i=0 ; i<count ; i++)
    {
        out[i] = LittleLong (in[i]);

        // Check bounds
        if (abs(out[i]) >= loadmodel->numedges)
        {
            Con_SafePrintf ("\x02Mod_LoadSurfedges: ");
            Con_SafePrintf ("edge in surfedge %d out of bounds (%d, max = %d) in %s\n", i, out[i], loadmodel->numedges, loadmodel->name);
            out[i] = MAX_MAP_EDGES; // OK?
        }
    }
}

/*
=================
Mod_LoadPlanes
=================
*/
void Mod_LoadPlanes (lump_t *l)
{
    int			i, j;
    mplane_t	*out;
    dplane_t 	*in;
    int			count;
    int			bits;

    in = (void *)(mod_base + l->fileofs);
    Mod_ChkLumpSize ("Mod_LoadPlanes", l, sizeof(*in));
    count = l->filelen / sizeof(*in);

    if (count > MAX_MAP_PLANES)
    {
        Con_SafePrintf ("\x02Mod_LoadPlanes: ");
        Con_Printf ("excessive planes (%d, normal max = %d) in %s\n", count, MAX_MAP_PLANES, loadmodel->name);
    }
    out = Heap_ZAlloc ( count*2*sizeof(*out));

    loadmodel->planes = out;
    loadmodel->numplanes = count;

    for ( i=0 ; i<count ; i++, in++, out++)
    {
        bits = 0;

        for (j=0 ; j<3 ; j++)
        {
            out->normal[j] = LittleFloat (in->normal[j]);

            if (out->normal[j] < 0)
            {
                bits |= 1<<j;
            }
        }
        out->dist = LittleFloat (in->dist);
        out->type = LittleLong (in->type);
        out->signbits = bits;
    }
}

/*
=================
Mod_RadiusFromBounds
=================
*/
float Mod_RadiusFromBounds (vec3_t mins, vec3_t maxs)
{
    int		i;
    vec3_t	corner;

    for (i=0 ; i<3 ; i++)
    {
        corner[i] = fabs(mins[i]) > fabs(maxs[i]) ? fabs(mins[i]) : fabs(maxs[i]);
    }
    return Length (corner);
}

/*
=================
Mod_CheckWorldMap

reliable detection of a world map.
=================
*/
qboolean Mod_CheckWorldMap (model_t *mod)
{
    int i;

    // not sure if the -11 is 1 too many, but it doesn't matter anyway
    for (i = 0; i < mod->numentities - 11; i++)
    {
        // catch an info_player_* entity
        // this is suitable for all gameplay types.
        if (mod->entities[i] == 'i' &&
                mod->entities[i + 1] == 'n' &&
                mod->entities[i + 2] == 'f' &&
                mod->entities[i + 3] == 'o' &&
                mod->entities[i + 4] == '_' &&
                mod->entities[i + 5] == 'p' &&
                mod->entities[i + 6] == 'l' &&
                mod->entities[i + 7] == 'a' &&
                mod->entities[i + 8] == 'y' &&
                mod->entities[i + 9] == 'e' &&
                mod->entities[i + 10] == 'r')
        {
            return true;
        }
    }
    return false;
}

/*
=================
Mod_LoadBrushModel
=================
*/
void Mod_LoadBrushModel (model_t *mod, void *buffer)
{
    int			i, j;
    int			mapcount;
    qboolean	worldmap;
    dheader_t	*header;
    dmodel_t 	*bm;

    loadmodel->type = mod_brush;

    header = (dheader_t *)buffer;

    // Check header is inside file buffer
    Mod_ChkFileSize ("Mod_LoadBrushModel", "header", 0, sizeof(dheader_t));

    i = LittleLong (header->version);

    if (i != BSPVERSION)
    {
        Sys_Error ("Mod_LoadBrushModel: %s has wrong version number (%i should be %i)", mod->name, i, BSPVERSION);
    }

    // swap all the lumps
    mod_base = (byte *)header;

    for (i = 0; i < (int) sizeof (dheader_t) / 4; i++)
    {
        ((int *)header)[i] = LittleLong (((int *)header)[i]);
    }

    // Check all lumps are OK and inside file buffer
    for (i = 0; i < HEADER_LUMPS; ++i)
    {
        Mod_ChkFileSize ("Mod_LoadBrushModel", LumpDesc[i], header->lumps[i].fileofs, header->lumps[i].filelen);
    }

    // load entities before anything else so we can see if it's the world
    Mod_LoadEntities (&header->lumps[LUMP_ENTITIES]);

    // find the worldspawn
    worldmap = Mod_CheckWorldMap (mod);

    // texture cache flushing.  to keep things clean, the texture cache is periodically flushed.
    // this, and the various safeguards i've built into the loading, should ensure it never
    // runs out.  of course, a dynamically allocated cache would be better, and i did have one
    // for a while, but management of it just became far too messy.
    for (mapcount = 0; mapcount < 20; mapcount++)
    {
        if (worldmap && mapcount > 10)
        {
            // flush every 10 frame or so
            GL_FlushTextureCache ();
            break; // valid break.
        }
        else if (!strcmp (loadmodel->name, "maps/start.bsp"))
        {
            // flush on start map load
            GL_FlushTextureCache ();
            break; // valid break.
        }
    }

    // now load the rest
    Mod_LoadVertexes (&header->lumps[LUMP_VERTEXES]);
    Mod_LoadEdges (&header->lumps[LUMP_EDGES]);
    Mod_LoadSurfedges (&header->lumps[LUMP_SURFEDGES]);
    Mod_LoadTextures (&header->lumps[LUMP_TEXTURES]);
    Mod_LoadLighting (&header->lumps[LUMP_LIGHTING]);
    Mod_LoadPlanes (&header->lumps[LUMP_PLANES]);
    Mod_LoadTexinfo (&header->lumps[LUMP_TEXINFO]);
    Mod_LoadFaces (&header->lumps[LUMP_FACES]);
    Mod_LoadMarksurfaces (&header->lumps[LUMP_MARKSURFACES]);
    Mod_LoadLeafs (&header->lumps[LUMP_LEAFS], &header->lumps[LUMP_VISIBILITY]);
    Mod_LoadNodes (&header->lumps[LUMP_NODES]);
    Mod_LoadClipnodes (&header->lumps[LUMP_CLIPNODES]);
    Mod_LoadSubmodels (&header->lumps[LUMP_MODELS]);

    Mod_MakeHull0 ();

    mod->numframes = 2;		// regular and alternate animation

    if (worldmap)
    {
        // made this a developer print now that i know it works...
        Con_DPrintf ("Carving effects from %s\n", mod->name);

        // world volume carving
        MH_WorldVolumeCarve (mod);
    }

    // set up the submodels (FIXME: this is confusing)
    for (i=0 ; i<mod->numsubmodels ; i++)
    {
        bm = &mod->submodels[i];
        mod->hulls[0].firstclipnode = bm->headnode[0];

        for (j=1 ; j<MAX_MAP_HULLS ; j++)
        {
            mod->hulls[j].firstclipnode = bm->headnode[j];
            mod->hulls[j].lastclipnode = mod->numclipnodes-1;
        }
        mod->firstmodelsurface = bm->firstface;
        mod->nummodelsurfaces = bm->numfaces;

        VectorCopy (bm->maxs, mod->maxs);
        VectorCopy (bm->mins, mod->mins);

        mod->radius = Mod_RadiusFromBounds (mod->mins, mod->maxs);

        mod->numleafs = bm->visleafs;

        if (i < mod->numsubmodels - 1)
        {
            // duplicate the basic information
            char	name[10];

            va_snprintf ("Mod_LoadBrushModel", name, sizeof(name), "*%d", i + 1);
            loadmodel = Mod_FindName (name);
            *loadmodel = *mod;
            strlcpy_s (loadmodel->name, name);
            mod = loadmodel;
        }
    }
}

/*
==============================================================================

ALIAS MODELS

==============================================================================
*/

aliashdr_t	*pheader;

stvert_t	stverts[MAXALIASVERTS];
mtriangle_t	triangles[MAXALIASTRIS];

// a pose is a single set of vertexes.  a frame may be
// an animating sequence of poses
trivertx_t	*poseverts[MAXALIASFRAMES];
int			posenum;

/*
=================
Mod_LoadAliasFrame
=================
*/
void *Mod_LoadAliasFrame (void *pin, maliasframedesc_t *frame)
{
    int				i;
    trivertx_t		*pinframe;
    daliasframe_t	*pdaliasframe;

    pdaliasframe = (daliasframe_t *)pin;

    // Check frame is inside file buffer after the fall fails brutally here :(
    Mod_ChkFileSize ("Mod_LoadAliasFrame", "frame", (byte *)pdaliasframe - mod_base, sizeof(daliasframe_t));

    // heavy checking
    strlcpy (frame->name, pdaliasframe->name, sizeof(frame->name));
    frame->firstpose = posenum;
    frame->numposes = 1;

    for (i = 0; i < 3; i++)
    {
        // these are byte values, so we don't have to worry about endianness
        frame->bboxmin[i] = pdaliasframe->bboxmin.v[i] * pheader->scale[i] + pheader->scale_origin[i];
        frame->bboxmax[i] = pdaliasframe->bboxmax.v[i] * pheader->scale[i] + pheader->scale_origin[i];
    }
    frame->radius = Mod_RadiusFromBounds (frame->bboxmin, frame->bboxmax);

    pinframe = (trivertx_t *)(pdaliasframe + 1);

    // Check trivertex is inside file buffer
    Mod_ChkFileSize ("Mod_LoadAliasFrame", "trivertex", (byte *)pinframe - mod_base, sizeof(trivertx_t));

    if (posenum >= MAXALIASFRAMES)
    {
        Host_Error ("Mod_LoadAliasFrame: invalid # of frames (%d, max = %d) in %s", posenum, MAXALIASFRAMES, loadmodel->name);
    }
    poseverts[posenum] = pinframe;
    posenum++;

    pinframe += pheader->numverts;

    return (void *)pinframe;
}


/*
=================
Mod_LoadAliasGroup
=================
*/
void *Mod_LoadAliasGroup (void * pin,  maliasframedesc_t *frame)
{
    daliasgroup_t		*pingroup;
    int					i, numframes;
    daliasinterval_t	*pin_intervals;
    void				*ptemp;

    pingroup = (daliasgroup_t *)pin;

    // Check group is inside file buffer
    Mod_ChkFileSize ("Mod_LoadAliasGroup", "group", (byte *)pingroup - mod_base, sizeof(daliasgroup_t));

    numframes = LittleLong (pingroup->numframes);

    frame->firstpose = posenum;
    frame->numposes = numframes;

    for (i=0 ; i<3 ; i++)
    {
        // these are byte values, so we don't have to worry about endianness
        frame->bboxmin[i] = pingroup->bboxmin.v[i] * pheader->scale[i] + pheader->scale_origin[i];
        frame->bboxmax[i] = pingroup->bboxmax.v[i] * pheader->scale[i] + pheader->scale_origin[i];
    }
    frame->radius = Mod_RadiusFromBounds (frame->bboxmin, frame->bboxmax);

    pin_intervals = (daliasinterval_t *)(pingroup + 1);

    // Check intervals are inside file buffer
    Mod_ChkFileSize ("Mod_LoadAliasGroup", "intervals", (byte *)pin_intervals - mod_base, numframes * sizeof(daliasinterval_t));

    frame->interval = LittleFloat (pin_intervals->interval);

    pin_intervals += numframes;

    ptemp = (void *)pin_intervals;

    for (i=0 ; i<numframes ; i++)
    {
        // Check trivertex is inside file buffer
        Mod_ChkFileSize ("Mod_LoadAliasGroup", "trivertex", (byte *)((daliasframe_t *)ptemp + 1) - mod_base, sizeof(trivertx_t));

        if (posenum >= MAXALIASFRAMES)
        {
            Host_Error ("Mod_LoadAliasGroup: invalid # of frames (%d, max = %d) in %s", posenum, MAXALIASFRAMES, loadmodel->name);
        }
        poseverts[posenum] = (trivertx_t *)((daliasframe_t *)ptemp + 1);
        posenum++;

        ptemp = (trivertx_t *)((daliasframe_t *)ptemp + 1) + pheader->numverts;
    }
    return ptemp;
}

//=========================================================

/*
=================
Mod_FloodFillSkin

Fill background pixels so mipmapping doesn't have haloes - Ed
=================
*/

typedef struct
{
    short		x, y;
} floodfill_t;

// must be a power of 2
#define FLOODFILL_FIFO_SIZE 0x1000
#define FLOODFILL_FIFO_MASK (FLOODFILL_FIFO_SIZE - 1)

#define FLOODFILL_STEP( off, dx, dy )						\
{															\
	if (pos[off] == fillcolor)								\
	{														\
		pos[off] = 255;										\
		fifo[inpt].x = x + (dx), fifo[inpt].y = y + (dy);	\
		inpt = (inpt + 1) & FLOODFILL_FIFO_MASK;			\
	}														\
	else if (pos[off] != 255) fdc = pos[off];				\
}

void Mod_FloodFillSkin( byte *skin, int skinwidth, int skinheight )
{
    byte				fillcolor = *skin; // assume this is the pixel to fill
    floodfill_t			fifo[FLOODFILL_FIFO_SIZE];
    int					inpt = 0, outpt = 0;
    int					filledcolor = -1;
    int					i;

    if (filledcolor == -1)
    {
        filledcolor = 0;

        // attempt to find opaque black
        for (i = 0; i < 256; ++i)
        {
            if (d_8to24table[i] == (255 << 0)) // alpha 1.0
            {
                filledcolor = i;
                break;
            }
        }
    }

    // can't fill to filled color or to transparent color (used as visited marker)
    if ((fillcolor == filledcolor) || (fillcolor == 255))
    {
        Con_DPrintf( "not filling skin from %d to %d\n", fillcolor, filledcolor );
        return;
    }
    fifo[inpt].x = 0, fifo[inpt].y = 0;
    inpt = (inpt + 1) & FLOODFILL_FIFO_MASK;

    while (outpt != inpt)
    {
        int			x = fifo[outpt].x, y = fifo[outpt].y;
        int			fdc = filledcolor;
        byte		*pos = &skin[x + skinwidth * y];

        outpt = (outpt + 1) & FLOODFILL_FIFO_MASK;

        if (x > 0)				FLOODFILL_STEP( -1, -1, 0 );
        if (x < skinwidth - 1)	FLOODFILL_STEP( 1, 1, 0 );
        if (y > 0)				FLOODFILL_STEP( -skinwidth, 0, -1 );
        if (y < skinheight - 1)	FLOODFILL_STEP( skinwidth, 0, 1 );
        skin[x + skinwidth * y] = fdc;
    }
}

/*
===============
Mod_LoadAllSkins
===============
*/
void *Mod_LoadAllSkins (int numskins, daliasskintype_t *pskintype)
{
    int						i, j, k, s;
    char					name[32];
    byte					*skin;
    byte					*texels;
    daliasskingroup_t		*pinskingroup;
    int						groupskins;
    daliasskininterval_t	*pinskinintervals;
    char					loadmodelname[32];

    skin = (byte *)(pskintype + 1);

    if (numskins < 1 || numskins > MAX_SKINS)
    {
        Sys_Error ("Mod_LoadAliasModel: Invalid # of skins: %d\n", numskins);
    }
    s = pheader->skinwidth * pheader->skinheight;

    // make the base name by ripping out the extension
    for (i = 0; ; i++)
    {
        loadmodelname[i] = loadmodel->name[i];

        if (loadmodelname[i] == '.')
        {
            loadmodelname[i] = 0;
        }

        if (loadmodelname[i] == 0) break;
    }

    for (i = 0; i < numskins; i++)
    {
        char Str[256];

        skin = (byte *)(pskintype + 1);

        sprintf (Str, "skin %d", i);
        Mod_ChkFileSize ("Mod_LoadAllSkins", Str, (byte *)pskintype - mod_base, sizeof(daliasskintype_t));
        Mod_ChkType ("Mod_LoadAllSkins", "skintype", Str, pskintype->type, ALIAS_SKIN_SINGLE, ALIAS_SKIN_GROUP);

        if (pskintype->type == ALIAS_SKIN_SINGLE)
        {
            Mod_FloodFillSkin (skin, pheader->skinwidth, pheader->skinheight);

            texels = Heap_ZAlloc (s);
            pheader->texels[i] = texels - (byte *) pheader;
            memcpy (texels, (byte *) (pskintype + 1), s);

            sprintf (name, "%s_%i", loadmodelname, i);

            pheader->gl_texture[i][0] =
                pheader->gl_texture[i][1] =
                    pheader->gl_texture[i][2] =
                        pheader->gl_texture[i][3] =	GL_LoadQuakeTexture (NULL, name, (byte *) (pskintype + 1), pheader->skinwidth, pheader->skinheight, (TEX_MIPMAP | TEX_MDL | TEX_MAKENORMAL));

            pskintype = (daliasskintype_t *) ((byte *) (pskintype + 1) + s);
        }
        else
        {
            // animating skin group.  yuck.
            pskintype++;
            pinskingroup = (daliasskingroup_t *) pskintype;

            sprintf (Str, "group %d", i);
            Mod_ChkFileSize ("Mod_LoadAllSkins", Str, (byte *)pinskingroup - mod_base, sizeof(daliasskingroup_t));

            groupskins = LittleLong (pinskingroup->numskins);
            pinskinintervals = (daliasskininterval_t *) (pinskingroup + 1);

            pskintype = (void *)(pinskinintervals + groupskins);

            for (j = 0; j < groupskins; j++)
            {
                sprintf (Str, "group %d, skin %d", i, j);
                Mod_ChkFileSize ("Mod_LoadAllSkins", Str, (byte *)pskintype - mod_base, s);

                Mod_FloodFillSkin (skin, pheader->skinwidth, pheader->skinheight);

                if (j == 0)
                {
                    texels = Heap_ZAlloc (s);
                    pheader->texels[i] = texels - (byte *) pheader;
                    memcpy (texels, (byte *) (pskintype), s);
                }
                sprintf (name, "%s_%i_%i", loadmodelname, i, j);

                pheader->gl_texture[i][j & 3] = GL_LoadQuakeTexture (NULL, name, (byte *) (pskintype), pheader->skinwidth, pheader->skinheight, (TEX_MIPMAP | TEX_MDL | TEX_MAKENORMAL));

                pskintype = (daliasskintype_t *) ((byte *) (pskintype) + s);
            }

            for (k = j; j < 4; j++)
            {
                pheader->gl_texture[i][j & 3] = pheader->gl_texture[i][j - k];
            }
        }
    }
    return (void *) pskintype;
}

//=========================================================================

/*
=================
Mod_LoadAliasModel
=================
*/
void Mod_LoadAliasModel (model_t *mod, void *buffer)
{
    int					i, j;
    int					size;
    int					version, numframes;
    mdl_t				*pinmodel;
    stvert_t			*pinstverts;
    dtriangle_t			*pintriangles;
    daliasframetype_t	*pframetype;
    daliasskintype_t	*pskintype;

    mod_base = (byte *)buffer;

    pinmodel = (mdl_t *) buffer;

    // Check version + header is inside file buffer
    Mod_ChkFileSize ("Mod_LoadAliasModel", "model header", 0, sizeof(int) + sizeof(mdl_t));

    version = LittleLong (pinmodel->version);

    if (version != ALIAS_VERSION)
    {
        Sys_Error ("%s has wrong version number (%i should be %i)", mod->name, version, ALIAS_VERSION);
    }

    // allocate space for a working header, plus all the data except the frames,
    // skin and group info
    size = 	sizeof (aliashdr_t) + (LittleLong (pinmodel->numframes) - 1) * sizeof (pheader->frames[0]);

    pheader = Heap_ZAlloc (size);

    mod->flags = LittleLong (pinmodel->flags);

    // endian-adjust and copy the data, starting with the alias model header
    pheader->boundingradius = LittleFloat (pinmodel->boundingradius);
    pheader->numskins = LittleLong (pinmodel->numskins);
    pheader->skinwidth = LittleLong (pinmodel->skinwidth);
    pheader->skinheight = LittleLong (pinmodel->skinheight);

    if (pheader->skinheight > MAX_LBM_HEIGHT)
    {
        Sys_Error ("model %s has a skin taller than %d", mod->name, MAX_LBM_HEIGHT);
    }
    pheader->numverts = LittleLong (pinmodel->numverts);

    if (pheader->numverts <= 0)
    {
        Sys_Error ("model %s has no vertices", mod->name);
    }

    if (pheader->numverts > MAXALIASVERTS)
    {
        Sys_Error ("model %s has too many vertices", mod->name);
    }
    pheader->numtris = LittleLong (pinmodel->numtris);

    if (pheader->numtris <= 0)
    {
        Sys_Error ("model %s has no triangles", mod->name);
    }

    if (pheader->numtris > MAXALIASTRIS)
    {
        Sys_Error ("model %s has too many triangles", mod->name);
    }
    pheader->numframes = LittleLong (pinmodel->numframes);
    numframes = pheader->numframes;

    if (numframes < 1)
    {
        Sys_Error ("Mod_LoadAliasModel: Invalid # of frames: %d\n", numframes);
    }
    pheader->size = LittleFloat (pinmodel->size) * ALIAS_BASE_SIZE_RATIO;
    mod->synctype = LittleLong (pinmodel->synctype);
    mod->numframes = pheader->numframes;

    for (i = 0; i < 3; i++)
    {
        pheader->scale[i] = LittleFloat (pinmodel->scale[i]);
        pheader->scale_origin[i] = LittleFloat (pinmodel->scale_origin[i]);
        pheader->eyeposition[i] = LittleFloat (pinmodel->eyeposition[i]);
    }

    // double size of eyes, since they are really hard to see in gl
    if (!strcasecmp (mod->name, "progs/eyes.mdl"))
    {
        VectorScale(pheader->scale, 2, pheader->scale);
        VectorScale(pheader->scale_origin, -30, pheader->scale_origin);
    }

    // load the skins
    pskintype = (daliasskintype_t *) &pinmodel[1];
    pskintype = Mod_LoadAllSkins (pheader->numskins, pskintype);

    // load base s and t vertices
    pinstverts = (stvert_t *) pskintype;

    // Check vertices are inside file buffer
    Mod_ChkFileSize ("Mod_LoadAliasModel", "vertices", (byte *)pinstverts - mod_base, pheader->numverts * sizeof(stvert_t));

    for (i = 0; i < pheader->numverts; i++)
    {
        stverts[i].onseam = LittleLong (pinstverts[i].onseam);
        stverts[i].s = LittleLong (pinstverts[i].s);
        stverts[i].t = LittleLong (pinstverts[i].t);
    }

    // load triangle lists
    pintriangles = (dtriangle_t *)&pinstverts[pheader->numverts];

    // Check triangles are inside file buffer
    Mod_ChkFileSize ("Mod_LoadAliasModel", "triangles", (byte *)pintriangles - mod_base, pheader->numtris * sizeof(dtriangle_t));

    for (i = 0; i < pheader->numtris; i++)
    {
        triangles[i].facesfront = LittleLong (pintriangles[i].facesfront);

        for (j = 0; j < 3; j++)
        {
            triangles[i].vertindex[j] = LittleLong (pintriangles[i].vertindex[j]);
        }
    }

    // load the frames
    posenum = 0;
    pframetype = (daliasframetype_t *)&pintriangles[pheader->numtris];

    mod->mins[0] = mod->mins[1] = mod->mins[2] = 255;
    mod->maxs[0] = mod->maxs[1] = mod->maxs[2] = 0;

    for (i = 0; i < numframes; i++)
    {
        aliasframetype_t	frametype;
        char				Str[256];

        sprintf (Str, "frame %d", i);
        Mod_ChkFileSize ("Mod_LoadAliasModel", Str, (byte *)pframetype - mod_base, sizeof(aliasframetype_t));

        frametype = LittleLong (pframetype->type);

        Mod_ChkType ("Mod_LoadAliasModel", "frametype", Str, frametype, ALIAS_SINGLE, ALIAS_GROUP);

        if (frametype == ALIAS_SINGLE)
        {
            pframetype = (daliasframetype_t *) Mod_LoadAliasFrame (pframetype + 1, &pheader->frames[i]);
        }
        else
        {
            pframetype = (daliasframetype_t *) Mod_LoadAliasGroup (pframetype + 1, &pheader->frames[i]);
        }

        for (j = 0; j < 3; j++)
        {
            mod->mins[j] = min (mod->mins[j], pheader->frames[i].bboxmin[j]);
            mod->maxs[j] = max (mod->maxs[j], pheader->frames[i].bboxmax[j]);
        }
    }
    mod->radius = Mod_RadiusFromBounds (mod->mins, mod->maxs);

    pheader->numposes = posenum;

    mod->type = mod_alias;

    // set the midpoint
    for (i = 0; i < 3; i++)
    {
        // add half the size to mins rather than taking a straight average cos the mins might be negative
        mod->mids[i] = mod->mins[i] + ((mod->maxs[i] - mod->mins[i]) / 2.0f);
    }
    mod->radius = Mod_RadiusFromBounds (mod->mins, mod->maxs);

    // build the draw lists
    GL_MakeAliasModelDisplayLists (mod, pheader);

    // set special effects - default to everything off
    pheader->bolt = false;
    pheader->flicker = false;
    pheader->glow = false;
    pheader->interpolate = false;
    pheader->missile = false;
    pheader->powerup = false;
    pheader->pulse = false;
    pheader->shambler = false;
    pheader->torch = false;
    pheader->translevel = 1;
    pheader->radius = 0;

    if (!strncmp (mod->name, "progs/flame", 11))
    {
        pheader->glow = true;
        pheader->flicker = true;
        pheader->torch = true;
        pheader->colour = 107;
        pheader->radius = 30;
    }
    else if (!strcmp(mod->name, "progs/flame0.mdl"))
    {
        pheader->glow = true;
        pheader->flicker = true;
        pheader->torch = true;
        pheader->colour = 107;
        pheader->radius = 40;
    }
    else if (!strcmp(mod->name, "progs/flame2.mdl"))
    {
        pheader->glow = true;
        pheader->flicker = true;
        pheader->torch = true;
        pheader->colour = 107;
        pheader->radius = 30;
    }
    else if (!strncmp(mod->name, "progs/bolt",10))
    {
        pheader->glow = true;
        pheader->bolt = true;
        pheader->colour = 246;
        pheader->radius = 15;
    }
    else if (!strcmp  (mod->name, "progs/missile.mdl"))
    {
        pheader->glow = true;
        pheader->flicker = true;
        pheader->missile = true;
        pheader->colour = 107;
        pheader->radius = 30;
    }
    else if (!strcmp  (mod->name, "progs/candle.mdl"))
    {
        pheader->glow = true;
        pheader->flicker = true;
        pheader->torch = true;
        pheader->colour = 107;
        pheader->radius = 30;
    }
    else if (!strcmp  (mod->name, "progs/lantern.mdl"))
    {
        pheader->glow = true;
        pheader->flicker = true;
        pheader->torch = true;
        pheader->colour = 107;
        pheader->radius = 30;
    }
    else if (!strcmp  (mod->name, "progs/s_light.mdl"))
    {
        pheader->glow = true;
        pheader->shambler = true;
        pheader->radius = 150;
        pheader->colour = 246;
    }
    else if (!strcmp (mod->name, "progs/quaddama.mdl"))
    {
        pheader->glow = true;
        pheader->flicker = true;
        pheader->powerup = true;
        pheader->radius = 100;
        pheader->colour = 246;
    }
    else if (!strcmp (mod->name, "progs/invulner.mdl"))
    {
        pheader->glow = true;
        pheader->flicker = true;
        pheader->powerup = true;
        pheader->radius = 100;
        pheader->colour = 100;
    }

    // leave the model on the hunk
    mod->extradata = pheader;
}

//=============================================================================

/*
=================
Mod_LoadSpriteFrame
=================
*/
void *Mod_LoadSpriteFrame (void * pin, mspriteframe_t **ppframe, int framenum)
{
    dspriteframe_t		*pinframe;
    mspriteframe_t		*pspriteframe;
    int					width, height, size, origin[2];
    char				name[64], str[256];

    pinframe = (dspriteframe_t *)pin;

    // Check sprite frame is inside file buffer
    sprintf (str, "frame %d", framenum);
    Mod_ChkFileSize ("Mod_LoadSpriteFrame", str, (byte *)pinframe - mod_base, sizeof(dspriteframe_t));

    width = LittleLong (pinframe->width);
    height = LittleLong (pinframe->height);
    size = width * height;

    Mod_ChkFileSize ("Mod_LoadSpriteFrame", str, (byte *)(pinframe + 1) - mod_base, size);

    pspriteframe = Heap_ZAlloc (sizeof (mspriteframe_t));

    memset (pspriteframe, 0, sizeof (mspriteframe_t));

    *ppframe = pspriteframe;

    pspriteframe->width = width;
    pspriteframe->height = height;
    origin[0] = LittleLong (pinframe->origin[0]);
    origin[1] = LittleLong (pinframe->origin[1]);

    pspriteframe->up = origin[1];
    pspriteframe->down = origin[1] - height;
    pspriteframe->left = origin[0];
    pspriteframe->right = width + origin[0];

    sprintf (name, "%s_%i", loadmodel->name, framenum);
    pspriteframe->gl_texture = GL_LoadQuakeTexture (NULL, name, (byte *) (pinframe + 1), width, height, (TEX_TEXTURE));

    return (void *)((byte *)pinframe + sizeof (dspriteframe_t) + size);
}


/*
=================
Mod_LoadSpriteGroup
=================
*/
void *Mod_LoadSpriteGroup (void * pin, mspriteframe_t **ppframe, int framenum)
{
    dspritegroup_t		*pingroup;
    mspritegroup_t		*pspritegroup;
    int					i, numframes;
    dspriteinterval_t	*pin_intervals;
    float				*poutintervals;
    void				*ptemp;

    pingroup = (dspritegroup_t *)pin;

    numframes = LittleLong (pingroup->numframes);

    pspritegroup = Heap_ZAlloc (sizeof (mspritegroup_t) + (numframes - 1) * sizeof (pspritegroup->frames[0]));

    pspritegroup->numframes = numframes;

    *ppframe = (mspriteframe_t *)pspritegroup;

    pin_intervals = (dspriteinterval_t *)(pingroup + 1);

    poutintervals = Heap_ZAlloc (numframes * sizeof (float));

    pspritegroup->intervals = poutintervals;

    for (i=0 ; i<numframes ; i++)
    {
        *poutintervals = LittleFloat (pin_intervals->interval);

        if (*poutintervals <= 0.0)
        {
            Sys_Error ("Mod_LoadSpriteGroup: interval<=0");
        }
        poutintervals++;
        pin_intervals++;
    }
    ptemp = (void *)pin_intervals;

    for (i=0 ; i<numframes ; i++)
    {
        ptemp = Mod_LoadSpriteFrame (ptemp, &pspritegroup->frames[i], framenum * 100 + i);
    }
    return ptemp;
}


/*
=================
Mod_LoadSpriteModel
=================
*/
void Mod_LoadSpriteModel (model_t *mod, void *buffer)
{
    int					i;
    int					version;
    dsprite_t			*pin;
    msprite_t			*psprite;
    int					numframes;
    int					size;
    dspriteframetype_t	*pframetype;

    mod_base = (byte *)buffer;

    pin = (dsprite_t *)buffer;

    // Check version+header is inside file buffer
    Mod_ChkFileSize ("Mod_LoadSpriteModel", "sprite header", 0, sizeof(int) + sizeof(dsprite_t));

    version = LittleLong (pin->version);

    if (version != SPRITE_VERSION)
    {
        Sys_Error ("%s has wrong version number (%i should be %i)", mod->name, version, SPRITE_VERSION);
    }
    numframes = LittleLong (pin->numframes);

    size = sizeof (msprite_t) +	(numframes - 1) * sizeof (psprite->frames);

    psprite = Heap_ZAlloc (size);

    mod->extradata = psprite;

    psprite->type = LittleLong (pin->type);
    psprite->maxwidth = LittleLong (pin->width);
    psprite->maxheight = LittleLong (pin->height);
    psprite->beamlength = LittleFloat (pin->beamlength);
    mod->synctype = LittleLong (pin->synctype);
    psprite->numframes = numframes;

    mod->mins[0] = mod->mins[1] = -psprite->maxwidth/2;
    mod->maxs[0] = mod->maxs[1] = psprite->maxwidth/2;
    mod->mins[2] = -psprite->maxheight/2;
    mod->maxs[2] = psprite->maxheight/2;

    // check for sprite replacement.  even if it's replaced we'll load it anyway to keep
    // the cache clean and consistent at all times...
    if (!strcmp (mod->name, "progs/s_bubble.spr"))
    {
        // underwater bubbles
        psprite->replace = P_BUBBLE;
    }
    else if (!strcmp (mod->name, "progs/s_explod.spr"))
    {
        // explosions.  these aren't replaced directly in the sprite drawing routines, but
        // are flagged as replaced anyway so that they are not drawn
        psprite->replace = P_BLAST;
    }
    else if (!strcmp (mod->name, "progs/s_rain.spr"))
    {
        // rain
        psprite->replace = P_RAIN;
    }
    else if (!strcmp (mod->name, "progs/s_light.spr"))
    {
        // light globes
        psprite->replace = P_FLARE2;
    }
    else
    {
        psprite->replace = P_NONE;
    }

    // load the frames
    if (numframes < 1)
    {
        Sys_Error ("Mod_LoadSpriteModel: Invalid # of frames: %d\n", numframes);
    }
    mod->numframes = numframes;

    pframetype = (dspriteframetype_t *)(pin + 1);

    for (i=0 ; i<numframes ; i++)
    {
        spriteframetype_t	frametype;
        char				Str[256];

        sprintf (Str, "frame %d", i);
        Mod_ChkFileSize ("Mod_LoadSpriteModel", Str, (byte *)pframetype - mod_base, sizeof(spriteframetype_t));

        frametype = LittleLong (pframetype->type);
        psprite->frames[i].type = frametype;

        Mod_ChkType ("Mod_LoadSpriteModel", "frametype", Str, frametype, SPR_SINGLE, SPR_GROUP);

        if (frametype == SPR_SINGLE)
        {
            pframetype = (dspriteframetype_t *)	Mod_LoadSpriteFrame (pframetype + 1, &psprite->frames[i].frameptr, i);
        }
        else
        {
            pframetype = (dspriteframetype_t *)	Mod_LoadSpriteGroup (pframetype + 1, &psprite->frames[i].frameptr, i);
        }
    }
    mod->type = mod_sprite;
}

//=============================================================================

/*
================
Mod_Print
================
*/
void Mod_Print (void)
{
    int		i;
    model_t	*mod;

    Con_Printf ("Cached models:\n");

    for (i=0, mod=mod_known ; i < mod_numknown ; i++, mod++)
    {
        // only alias models go in the cache!!!
        if (mod->type != mod_alias) continue;

        Con_Printf ("%8p : %s\n", mod->extradata, mod->name);
    }
}
