/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_light.c

#include "gl_state.h"

gltexture_t GLLightmapTextures[MAX_LIGHTMAPS];

/*
==================
R_AnimateLight
==================
*/
void R_AnimateLight (void)
{
    int			j, k;
    float		l;
    int			flight;
    int			clight;
    float		lerpfrac;
    float		backlerp;

    // light animations
    // 'm' is normal light, 'a' is no light, 'z' is double bright
    flight = (int) floor (cl.time * 10);
    clight = (int) ceil (cl.time * 10);
    lerpfrac = (cl.time * 10) - flight;
    backlerp = 1.0f - lerpfrac;

    for (j = 0; j < MAX_LIGHTSTYLES; j++)
    {
        if (!cl_lightstyle[j].length)
        {
            // was 256, changed to 264 for consistency
            d_lightstylevalue[j] = 264;
            continue;
        }
        else if (cl_lightstyle[j].length == 1)
        {
            // single length style so don't bother interpolating
            d_lightstylevalue[j] = 22 * (cl_lightstyle[j].map[0] - 'a');
            continue;
        }

        // interpolate animating light
        // frame just gone
        k = flight % cl_lightstyle[j].length;
        k = cl_lightstyle[j].map[k] - 'a';
        l = (float) (k * 22) * backlerp;

        // upcoming frame
        k = clight % cl_lightstyle[j].length;
        k = cl_lightstyle[j].map[k] - 'a';
        l += (float) (k * 22) * lerpfrac;

        d_lightstylevalue[j] = (int)l;
    }
}


/*
=============================================================================

DYNAMIC LIGHTS

=============================================================================
*/

/*
=============
R_MarkLights
=============
*/
void R_MarkLights (dlight_t *light, int bit, mnode_t *node)
{
    mplane_t	*splitplane;
    float		dist, l, maxdist;
    msurface_t	*surf;
    int			i, j, s, t, sidebit;
    vec3_t		impact;

loc0:
    if (node->contents < 0)	return;

    splitplane = node->plane;
    dist = PlaneDiff(light->origin, splitplane);

    if (dist > light->radius)
    {
        node = node->children[0];
        goto loc0;
    }

    if (dist < -light->radius)
    {
        node = node->children[1];
        goto loc0;
    }
    maxdist = light->radius * light->radius;

    // mark the polygons
    surf = cl.worldmodel->surfaces + node->firstsurface;

    for (i=0 ; i<node->numsurfaces ; i++, surf++)
    {
        dist = PlaneDiff(light->origin, surf->plane);

        if (dist >= 0)
        {
            sidebit = 0;
        }
        else
        {
            sidebit = SURF_PLANEBACK;
        }

        if ((surf->flags & SURF_PLANEBACK) != sidebit) continue;

        for (j = 0 ; j < 3 ; j++)
        {
            impact[j] = light->origin[j] - surf->plane->normal[j]*dist;
        }

        // clamp center of light to corner and check brightness
        l = DotProduct(impact, surf->texinfo->vecs[0]) + surf->texinfo->vecs[0][3] - surf->texturemins[0];
        s = l + 0.5;
        s = bound(0, s, surf->extents[0]);
        s = l - s;
        l = DotProduct(impact, surf->texinfo->vecs[1]) + surf->texinfo->vecs[1][3] - surf->texturemins[1];
        t = l + 0.5;
        t = bound(0, t, surf->extents[1]);
        t = l - t;

        // compare to minimum light
        if ((s*s + t*t + dist*dist) < maxdist)
        {
            if (surf->dlightframe != r_framecount)
            {
                // not dynamic until now
                surf->dlightbits = bit;
                surf->dlightframe = r_framecount;
            }
            else
            {
                // already dynamic
                surf->dlightbits |= bit;
            }
        }
    }

    if (node->children[0]->contents >= 0)
    {
        R_MarkLights (light, bit, node->children[0]);
    }

    if (node->children[1]->contents >= 0)
    {
        R_MarkLights (light, bit, node->children[1]);
    }
}

/*
=============
R_PushDlights
=============
*/
void R_PushDlights (mnode_t *headnode)
{
    int			i;
    dlight_t	*l;

    l = cl_dlights;

    for (i = 0; i < MAX_DLIGHTS; i++, l++)
    {
        if (l->die < cl.time || !l->radius) continue;

        R_MarkLights (l, 1 << i, headnode);
    }
}


/*
=============================================================================

LIGHT SAMPLING

=============================================================================
*/

mplane_t		*lightplane;
vec3_t			lightspot;
vec3_t			pointcolour;

int RecursiveLightPoint (mnode_t *node, vec3_t start, vec3_t end)
{
    int			r;
    float		front, back, frac;
    int			side;
    mplane_t	*plane;
    vec3_t		mid;
    msurface_t	*surf;
    int			s, t, ds, dt;
    int			i;
    mtexinfo_t	*tex;
    byte		*lightmap;
    unsigned	scale;
    int			maps;

    // didn't hit anything
    if (node->contents < 0) return -1;

    // calculate mid point
    plane = node->plane;

    front = PlaneDiff(start, plane);
    back = PlaneDiff(end, plane);

    side = front < 0;

    if ((back < 0) == side)
    {
        return RecursiveLightPoint (node->children[side], start, end);
    }
    frac = front / (front - back);

    mid[0] = start[0] + (end[0] - start[0]) * frac;
    mid[1] = start[1] + (end[1] - start[1]) * frac;
    mid[2] = start[2] + (end[2] - start[2]) * frac;

    // go down front side
    r = RecursiveLightPoint (node->children[side], start, mid);

    // hit something
    if (r >= 0) return r;

    // didn't hit anything
    if ((back < 0) == side) return -1;

    // check for impact on this node
    VectorCopy (mid, lightspot);

    surf = cl.worldmodel->surfaces + node->firstsurface;

    for (i = 0; i < node->numsurfaces; i++, surf++)
    {
        // lightmaps sky and water do no have dynamic light
        if (surf->flags & (SURF_DRAWTILED | SURF_DRAWTURB | SURF_DRAWSKY)) continue;

        tex = surf->texinfo;

        s = DotProduct (mid, tex->vecs[0]) + tex->vecs[0][3];
        t = DotProduct (mid, tex->vecs[1]) + tex->vecs[1][3];;

        if (s < surf->texturemins[0] || t < surf->texturemins[1]) continue;

        ds = s - surf->texturemins[0];
        dt = t - surf->texturemins[1];

        if (ds > surf->extents[0] || dt > surf->extents[1]) continue;

        ds >>= 4;
        dt >>= 4;

        lightmap = surf->samples;

        // set to all black
        pointcolour[0] = pointcolour[1] = pointcolour[2] = 0;

        if (lightmap)
        {
            lightmap += 3 * (dt * ((surf->extents[0]>>4)+1) + ds);

            for (maps = 0; maps < MAXLIGHTMAPS && surf->styles[maps] != 255; maps++)
            {
                scale = d_lightstylevalue[surf->styles[maps]];

                pointcolour[0] += lightmap[0] * scale * ONEOVER255;
                pointcolour[1] += lightmap[1] * scale * ONEOVER255;
                pointcolour[2] += lightmap[2] * scale * ONEOVER255;

                lightmap += 3 * (((surf->extents[0]>>4)+1) * ((surf->extents[1]>>4)+1));
            }
        }

        // store the plane so we know if we've hit something
        lightplane = surf->plane;

        return 1;
    }

    // go down back side
    return RecursiveLightPoint (node->children[!side], mid, end);
}

void R_LightPoint (vec3_t p, vec3_t colour)
{
    vec3_t		start;
    vec3_t		end;
    int			r;
    int			lnum;
    vec3_t		dist;
    float		add;
    int			x;
    model_t		*mod;
    model_t		*shortest;

    if (!cl.worldmodel->lightdata)
    {
        colour[0] = 255;
        colour[1] = 255;
        colour[2] = 255;
        return;
    }

    // set up end point
    // this must be done seperately for each model checked as it's passed as a pointer to
    // RecursiveLightPoint, and therefore may have it's value changed
    end[0] = p[0];
    end[1] = p[1];
    end[2] = p[2] - 2048;

    // nothing hit yet
    lightplane = NULL;

    // do the world first
    r = RecursiveLightPoint (cl.worldmodel->nodes, p, end);

    // did we hit?
    if (lightplane)
    {
        // copy out the data
        cl.worldmodel->lightplane = lightplane;

        cl.worldmodel->lightspot[0] = lightspot[0];
        cl.worldmodel->lightspot[1] = lightspot[1];
        cl.worldmodel->lightspot[2] = lightspot[2];

        cl.worldmodel->pointcolour[0] = pointcolour[0];
        cl.worldmodel->pointcolour[1] = pointcolour[1];
        cl.worldmodel->pointcolour[2] = pointcolour[2];
    }
    else
    {
        // set to valid data for shortest path testing
        cl.worldmodel->lightplane = NULL;
        cl.worldmodel->lightspot[2] = -1000000;

        cl.worldmodel->pointcolour[0] = 0;
        cl.worldmodel->pointcolour[1] = 0;
        cl.worldmodel->pointcolour[2] = 0;
    }

    // find the shortest path - this will be the model with the highest value in lightspot[2]
    // init it to the worldmodel first
    shortest = cl.worldmodel;

    // now check the brush models
    for (x = 0; x < cl_numvisedicts; x++)
    {
        mod = cl_visedicts[x]->model;

        if (mod->type != mod_brush) continue;

        if (mod->name[0] != '*') continue;

        // set up start point
        // offset the start [0] and [1] points by entity origin to do func_train bmodels
        // this is positioned above the model to account for moving bmodels (note the "-" in [2])
        start[0] = p[0] + cl_visedicts[x]->origin[0];
        start[1] = p[1] + cl_visedicts[x]->origin[1];
        start[2] = p[2] - cl_visedicts[x]->origin[2];

        // set up end point
        // this must be done seperately for each model checked as it's passed as a pointer to
        // RecursiveLightPoint, and therefore may have it's value changed
        // offset the end [0] and [1] points by entity origin to do func_train bmodels
        end[0] = p[0] + cl_visedicts[x]->origin[0];
        end[1] = p[1] + cl_visedicts[x]->origin[1];
        end[2] = p[2] - 2048 + cl_visedicts[x]->origin[2];

        // NULL it again
        lightplane = NULL;

        // and do it
        r = RecursiveLightPoint (mod->nodes + mod->hulls[0].firstclipnode, start, end);

        // did we hit?
        if (lightplane)
        {
            // copy out the data
            mod->lightplane = lightplane;

            mod->lightspot[0] = lightspot[0];
            mod->lightspot[1] = lightspot[1];

            // store the colour for this model.  it hit but we don't know if it's the shortest
            // yet, so each model must store it's own colour
            mod->pointcolour[0] = pointcolour[0];
            mod->pointcolour[1] = pointcolour[1];
            mod->pointcolour[2] = pointcolour[2];

            // add origin[2] cos the bmodel may move!!!
            // it's official.  i'm a stupid twit.  i was using the origin of the *model* originally,
            // not of the *entity*.  this works just fine.
            mod->lightspot[2] = lightspot[2] + cl_visedicts[x]->origin[2];

            // check for shortest path
            if (mod->lightspot[2] > shortest->lightspot[2])
            {
                shortest = mod;
            }
        }
        else
        {
            mod->lightplane = NULL;

            mod->pointcolour[0] = 0;
            mod->pointcolour[1] = 0;
            mod->pointcolour[2] = 0;
        }
    }

    // copy the shortest out to the globals
    lightplane = shortest->lightplane;

    lightspot[0] = shortest->lightspot[0];
    lightspot[1] = shortest->lightspot[1];
    lightspot[2] = shortest->lightspot[2];

    // store pointcolour back to colour
    if (!lightplane)
    {
        // hit nothing
        colour[0] = 0;
        colour[1] = 0;
        colour[2] = 0;
    }
    else
    {
        // take the pointcolour from the shortest
        // (previous Realms took the global pointcolour, which may or may not have been the shortest)
        colour[0] = shortest->pointcolour[0];
        colour[1] = shortest->pointcolour[1];
        colour[2] = shortest->pointcolour[2];
    }

    // and add dynamic lights
    for (lnum = 0; lnum < MAX_DLIGHTS; lnum++)
    {
        if (cl_dlights[lnum].die >= cl.time)
        {
            VectorSubtract (p, cl_dlights[lnum].origin, dist);

            add = cl_dlights[lnum].radius - Length (dist);

            if (add > 0)
            {
                colour[0] += (cl_dlights[lnum].RGB[0] * add);
                colour[1] += (cl_dlights[lnum].RGB[1] * add);
                colour[2] += (cl_dlights[lnum].RGB[2] * add);
            }
        }
    }
}


/*
=============================================================================

LIGHTMAPS

=============================================================================
*/

lightmapinfo_t *lightmapinfo[MAX_LIGHTMAPS];

#define CDIV  200
#define CMULT 200

byte BLTable[262144];
unsigned blcr[18 * 18];
unsigned blcg[18 * 18];
unsigned blcb[18 * 18];

void R_StoreLightmap (msurface_t *surf, lightmapinfo_t *lm)
{
    int stride = BLOCK_WIDTH * LIGHTMAP_BYTES - (surf->smax << 2);
    int i, j;
    int blofs;
    byte *dest = &lm->texels[surf->base];

    for (i = 0, blofs = 0; i < surf->tmax; i++, dest += stride)
    {
        for (j = 0; j < surf->smax; j++, blofs++)
        {
            // do our table lookup now
            // set to 255 if the value > the sizeof the the table
            // don;t bother filling in dest[3] cos it's unused
            dest[0] = blcr[blofs] > 262143 ? 255 : BLTable[blcr[blofs]];
            dest[1] = blcg[blofs] > 262143 ? 255 : BLTable[blcg[blofs]];
            dest[2] = blcb[blofs] > 262143 ? 255 : BLTable[blcb[blofs]];

            dest += LIGHTMAP_BYTES;
        }
    }

    // mark it as modified to force an upload when it's next seen
    lm->modified = true;
    lm->stainactive = false;
    lm->stained = true;
}


/*
===============
R_AddDynamicLights
===============
*/
void R_AddDynamicLights (msurface_t *surf)
{
    int			lnum;
    int			sd, td;
    float		dist, rad, minlight;
    vec3_t		impact, local;
    int			s, t;
    int			i;
    mtexinfo_t	*tex;
    int			blofs;
    float		ladd;
    dlight_t	*dl;

    tex = surf->texinfo;

    // set initial blocklights here so we can avoid a potentially
    // unnecessary run through R_BuildLightmap just to get them done
    for (i = 0; i < surf->smax * surf->tmax; i++)
    {
        blcr[i] = blcg[i] = blcb[i] = 0;
    }

    // now add in the dlights
    for (lnum = 0; lnum < MAX_DLIGHTS; lnum++)
    {
        // not lit by this light
        if (!(surf->dlightbits & (1 << lnum))) continue;

        // set the current dlight
        dl = &cl_dlights[lnum];

        // messy crap
        rad = dl->radius;
        dist = PlaneDiff(dl->origin, surf->plane);
        rad -= fabs (dist);
        minlight = dl->minlight;

        if (rad < minlight) continue;

        minlight = rad - minlight;

        for (i = 0; i < 3; i++)
        {
            impact[i] = dl->origin[i] - surf->plane->normal[i] * dist;
        }
        local[0] = DotProduct (impact, tex->vecs[0]) + tex->vecs[0][3];
        local[1] = DotProduct (impact, tex->vecs[1]) + tex->vecs[1][3];

        local[0] -= surf->texturemins[0];
        local[1] -= surf->texturemins[1];

        for (t = 0; t < surf->tmax; t++)
        {
            td = local[1] - t * 16;

            if (td < 0) td = -td;

            for (s = 0; s < surf->smax; s++)
            {
                sd = local[0] - s * 16;

                if (sd < 0)
                {
                    sd = -sd;
                }

                if (sd > td)
                {
                    dist = sd + (td >> 1);
                }
                else
                {
                    dist = td + (sd >> 1);
                }

                if (dist < minlight)
                {
                    // we know it's been hit by a dlight now
                    surf->cached_dlight = true;
                    surf->activedlightbits |= (1 << lnum);

                    blofs = t * surf->smax + s;
                    ladd = (rad - dist) * 256;

                    blcr[blofs] += ladd * dl->RGB[0];
                    blcg[blofs] += ladd * dl->RGB[1];
                    blcb[blofs] += ladd * dl->RGB[2];
                }
            }
        }
    }
}


/*
================
R_InitBLTable

used for scaling the lighting range and gives R_BuildLightmap a lookup table to work from.
To make things brighter increase CMULT above towards 255, to make them darker increase
CDIV towards 255.  Any other operation will probably have weird effects.  In theory
this table *should* be 4 (MAXLIGHTMAPS) * 255 (max value in lightmaps) * 550 (max that a
style can scale by), or 561000, so I'm taking a wee risk with the lower size, but I've
never seen Realm crash because of this.  Ideally if there was a short data type that
clamped rather than wrapped I'd use that for blocklights... sigh...
================
*/
void R_InitBLTable (void)
{
    int		i, j, x;
    float	k;
    float	MaxLight;
    int		lexp = 1;

    if (gl_dot3bump.value)
    {
        MaxLight = 255;
    }
    else
    {
        MaxLight = 255 / BUMPRESCALE;
    }

    for (i = 0; i < 262144; i++)
    {
        j = i >> 7;

        if (j > 255)
        {
            j = 255;
        }

        // scale the lighting
        // this drastically (dramatically) improves appearance!!!
        k = ((float) j) / 255; //CDIV;

        for (x = 0; x < lexp; x++)
        {
            k *= k;
            k *= 2;
            if (k > 1.0) k = 1.0;
        }

        // scale light intensity upwards if we're bumpmapping
        // (do it in reverse so that colours don't wash out)
        if (!gl_dot3bump.value)
        {
            k /= BUMPRESCALE;
        }
        j = k * 255; //CMULT;

        if (j > (int) MaxLight)
        {
            j = (int) MaxLight;
        }
        BLTable[i] = r_fullbright.value ? 255 : j;
    }
}


/*
===============
R_BuildLightMap

Combine and scale multiple lightmaps into the 8.8 format in blocklights
===============
*/
void R_BuildLightMap (msurface_t *surf, qboolean initBLs)
{
    int			i, j, size;
    byte		*lightmap;
    unsigned	scale;
    int			maps;

    size = surf->smax * surf->tmax;
    lightmap = surf->samples;

    if (!cl.worldmodel->lightdata)
    {
        // set to fullbright and get out
        for (i = 0; i < size; i++)
        {
            blcr[i] = blcg[i] = blcb[i] = 262143;
        }
        return;
    }

    // clear to no light
    // happens for everything except dynamic lights, where this is done in R_AddDynamicLights instead.
    // if there is no dynamic light then we've avoided an unnecessary run through here just to get a
    // base value in the blocklights
    if (initBLs)
    {
        for (i = 0; i < size; i++)
        {
            blcr[i] = blcg[i] = blcb[i] = 0;
        }
    }

    // now add the lightmaps
    if (lightmap)
    {
        for (maps = 0; maps < MAXLIGHTMAPS && surf->styles[maps] != 255; maps++)
        {
            scale = d_lightstylevalue[surf->styles[maps]];
            surf->cached_light[maps] = scale;	// 8.8 fraction

            // scale can never be 1 but can be 0...
            if (scale)
            {
                for (i = 0, j = 0; i < size; i++)
                {
                    blcr[i] += lightmap[j++] * scale;
                    blcg[i] += lightmap[j++] * scale;
                    blcb[i] += lightmap[j++] * scale;
                }
            }
            lightmap += size * 3;	// skip to next lightmap
        }
    }
}


void R_UploadSingleLightmap (lightmapinfo_t *lm)
{
    // just upload the full thing.  "theRect" never worked properly anyway...
    // lightmaps are uploaded as needed, rather than blindly uploading every one
    // that was modified, to avoid possibly having to upload a lightmap that
    // isn't in the draw list, and to reduce texture changes per frame
    glTexSubImage2D (GL_TEXTURE_2D,
                     0,
                     0,
                     0,
                     BLOCK_WIDTH,
                     BLOCK_HEIGHT,
                     GL_RGBA,
                     GL_UNSIGNED_BYTE,
                     lm->texels);

    // don;t let it catch the next surf
    lm->modified = false;
}

void R_UpdateLightmapTextures (void)
{
    int i;

    for (i = 0; i < MAX_LIGHTMAPS; i++)
    {
        // no more available lightmaps
        if (!lightmapinfo[i]) break;

        // has it been modified?
        if (lightmapinfo[i]->modified)
        {
            // bind the texture.  this will give us multiple binds, but blah.
            glBindTexture (GL_TEXTURE_2D, lightmapinfo[i]->gltexturenumber);

            // send through the uploader
            R_UploadSingleLightmap (lightmapinfo[i]);
        }
    }
}


void R_AnimateSurfaceLight (msurface_t *surf)
{
    int			maps;
    qboolean	dynamic;

    // no dlights yet
    surf->activedlightbits = 0;

    // easy dynamic checking
    dynamic = (surf->dlightframe == r_framecount && r_dynamic.value);

    // previously dynamic, but not now - restore the original lightmap
    // will also catch any lightstyle changes
    if (surf->cached_dlight && !dynamic)
    {
        // rebuild the lightmap
        R_BuildLightMap (surf, true);

        // store in the surfaces lightmap texture
        R_StoreLightmap (surf, surf->lightmap);

        // not dynamic any more
        surf->cached_dlight = false;

        // no need to check any further
        return;
    }

    // previously dynamic and marked as dynamic this frame, so do a rebuild even
    // if it isn't actually dynamic.  will also catch any lightstyle changes
    if (surf->cached_dlight && dynamic)
    {
        // add any dynamic lights which hit
        R_AddDynamicLights (surf);

        // rebuild the lightmap
        R_BuildLightMap (surf, false);

        // store in the surfaces lightmap texture
        // even if a dynamic didn;t hit (we want to restore the original lightmap still)
        R_StoreLightmap (surf, surf->lightmap);

        // no need to check any further
        // leave surf->cached_dlight alone here, R_AddDynamicLights will give it it's
        // correct value for the next frame
        return;
    }

    // check for dynamic light modification
    if (dynamic)
    {
        // no dlights yet
        // (a marked surf may not actually be hit by a dlight)
        surf->cached_dlight = false;

        // add dynamic lights
        R_AddDynamicLights (surf);

        // did it hit, or had it a lightstyle?
        if (surf->cached_dlight)
        {
            // build the surface lightmap (will also catch any lightstyles on the surf)
            R_BuildLightMap (surf, false);

            // store the data
            R_StoreLightmap (surf, surf->lightmap);

            // no more checking needed
            return;
        }
    }

    // lightstyle modification
    for (maps = 0; maps < MAXLIGHTMAPS && surf->styles[maps] != 255; maps++)
    {
        if (d_lightstylevalue[surf->styles[maps]] != surf->cached_light[maps])
        {
            // update the lightmap
            R_BuildLightMap (surf, true);

            // store in the surfaces lightmap texture
            R_StoreLightmap (surf, surf->lightmap);

            // no need to check any further
            return;
        }
    }
}


void R_ClearLightmaps (void)
{
    int i;

    // release any lightmaps left over from the previous map
    for (i = 0; i < MAX_LIGHTMAPS; i++)
    {
        // release lm memory
        if (lightmapinfo[i])
        {
            // delete the texture objects
            glDeleteTextures (1, (GLuint *)&lightmapinfo[i]->gltexturenumber);
            glDeleteTextures (1, (GLuint *)&lightmapinfo[i]->staintexturenumber);

            // release memory
            free (lightmapinfo[i]);

            // set to a valid pointer
            lightmapinfo[i] = NULL;
        }

        // clear down the texture struct
        memset (&GLLightmapTextures[i], 0, sizeof (gltexture_t));
    }
}


void GL_AllocateLightmap (int texnum, int baseintensity)
{
    // allocate memory for the lightmap
    lightmapinfo[texnum] = (lightmapinfo_t *) malloc (sizeof (lightmapinfo_t));

    // set base data
    memset (lightmapinfo[texnum]->allocated, 0, sizeof (lightmapinfo[texnum]->allocated));
    memset (lightmapinfo[texnum]->texels, baseintensity, sizeof (lightmapinfo[texnum]->texels));
    lightmapinfo[texnum]->num = texnum;

    // upload to OpenGL
    // all subsequent calls, including uploading the initial data set, will use glTexSubImage2D
    GL_LoadQuakeTexture (&GLLightmapTextures[texnum], "", lightmapinfo[texnum]->texels, BLOCK_WIDTH, BLOCK_HEIGHT, TEX_LIGHTMAP);

    // store out the texture numbers
    lightmapinfo[texnum]->gltexturenumber = GLLightmapTextures[texnum].gltexnum;
    lightmapinfo[texnum]->staintexturenumber = GLLightmapTextures[texnum].lumatexnum;

    // set it to explicitly modified so that Quake will build it for us when the map runs
    lightmapinfo[texnum]->modified = true;
}


// returns a texture number and the position inside it
lightmapinfo_t *LMAllocBlock (int w, int h, int *x, int *y)
{
    int		i, j;
    int		best, best2;
    int		texnum;

    for (texnum = 0; texnum < MAX_LIGHTMAPS; texnum++)
    {
        best = BLOCK_HEIGHT;

        // use dynamic lightmap memory allocation
        if (!lightmapinfo[texnum])
        {
            GL_AllocateLightmap (texnum, LM_FULLDARK);
        }

        for (i = 0; i < BLOCK_WIDTH - w; i++)
        {
            best2 = 0;

            for (j = 0; j < w; j++)
            {
                if (lightmapinfo[texnum]->allocated[i + j] >= best) break;
                if (lightmapinfo[texnum]->allocated[i + j] > best2) best2 = lightmapinfo[texnum]->allocated[i + j];
            }

            if (j == w)
            {
                // this is a valid spot
                *x = i;
                *y = best = best2;
            }
        }

        if (best + h > BLOCK_HEIGHT) continue;

        for (i = 0; i < w; i++)
        {
            lightmapinfo[texnum]->allocated[*x + i] = best + h;
        }
        return lightmapinfo[texnum];
    }

    // if we have > 1024 lightmaps we have other things to worry about anyway...
    Sys_Error ("You mave more than %i Lightmaps!!!", MAX_LIGHTMAPS);

    // keep the stupid compiler happy
    return NULL;
}


/*
========================
GL_CreateSurfaceLightmap
========================
*/
void GL_CreateSurfaceLightmap (msurface_t *surf)
{
    // fill in smax and tmax
    surf->smax = (surf->extents[0] >> 4) + 1;
    surf->tmax = (surf->extents[1] >> 4) + 1;

    // allocate a block in the lightmap texels.  also fills in light_s and light_t
    surf->lightmap = LMAllocBlock (surf->smax, surf->tmax, &surf->light_s, &surf->light_t);

    // set the offset in the lm data
    surf->base = (surf->light_t * BLOCK_WIDTH + surf->light_s) * LIGHTMAP_BYTES;

    // give dlight frame an invalid value to avoid building dlights
    surf->dlightframe = -1;

    // build the lightmap
    R_BuildLightMap (surf, true);

    // and store it
    R_StoreLightmap (surf, surf->lightmap);
}


void GL_RemakeSurfaceLightmaps (void)
{
    int		i, j;
    model_t	*m;

    // go through all of the models
    for (j = 1; j < MAX_MODELS; j++)
    {
        // get the current model
        m = cl.model_precache[j];

        // no more models
        if (!m) break;

        // don;t do * models (their data comes from the world)
        if (m->name[0] == '*') continue;

        // make sure we only get brush models
        if (m->type != mod_brush) continue;

        // now go through all of the surfs in the model
        for (i = 0; i < m->numsurfaces; i++)
        {
            // moved this check to first so that the redundant call to
            // GL_CreateSurfaceLightmap is avoided
            if (m->surfaces[i].flags & SURF_DRAWTURB) continue;
            if (m->surfaces[i].flags & SURF_DRAWSKY) continue;

            // recreate the lightmap
            R_BuildLightMap (&m->surfaces[i], true);

            // and store it
            R_StoreLightmap (&m->surfaces[i], m->surfaces[i].lightmap);
        }
    }
}


/*
========================================================================================================================================

						STAINMAPS

	These use a similar system and the same struct as lightmaps, so they go in here.

	The code was evolved from some work I did for Alex Shadowwalker, which in turn came from somewhere else I can't remember.
	This final system is a hybrid of the 2, and incorporates some of the previous Realm 8 system.

	It's a lot more sensible and intuitive than the old Realm system I'd cribbed from somebody's Q2 system.  Apologies for
	not giving full named credit, but I really don't remember.

========================================================================================================================================
*/

void R_StainSurf (msurface_t *surf, float *parms)
{
    mtexinfo_t		*tex;
    vec3_t			impact, local;
    int				sd, td, s, t;
    float			fdist, frad, fminlight, fsacc, ftacc;
    unsigned short	*bl;
    qboolean		stained;
    float			amm;

    // initially unstained cos it may not even be hit!!!
    stained = false;

    if (!surf->samples) return;

    tex = surf->texinfo;

    frad = *parms;

    // removed axial optimisations cos plane->type is relative to the world orientation, not the
    // point orientation
    fdist = PlaneDiff((parms + 1), surf->plane);
    frad -= fabs(fdist);
    fminlight = 0;
    if (frad < fminlight) return;	//not hit
    fminlight = frad - fminlight;

    // unwinding loops is good for your health
    impact[0] = (parms + 1)[0] - surf->plane->normal[0] * fdist;
    impact[1] = (parms + 1)[1] - surf->plane->normal[1] * fdist;
    impact[2] = (parms + 1)[2] - surf->plane->normal[2] * fdist;

    local[0] = DotProduct (impact, tex->vecs[0]) + tex->vecs[0][3] - surf->texturemins[0];
    local[1] = DotProduct (impact, tex->vecs[1]) + tex->vecs[1][3] - surf->texturemins[1];

    // use stainsamples here...
    bl = surf->stainsamples;

    for (t = 0, ftacc = 0; t < surf->tmax; t++, ftacc += 16)
    {
        td = local[1] - ftacc;

        if (td < 0) td = -td;

        for (s = 0, fsacc = 0; s < surf->smax; s++, fsacc += 16, bl += 3)
        {
            sd = local[0] - fsacc;

            if (sd < 0) sd = -sd;

            fdist = (sd > td) ? sd + (td >> 1) : td + (sd >> 1);

            if (fdist < fminlight)
            {
                amm = (frad - fdist);

                // note: this could in theory wrap!!!
                // it's unlikely as parms are floats < 1
                bl[0] += amm * parms[4];
                bl[1] += amm * parms[5];
                bl[2] += amm * parms[6];

                // only mark the surf as stained if it actually *has* been...
                stained = true;
            }
        }
    }

    if (stained)
    {
        surf->lightmap->stained = true;
        surf->lightmap->stainactive = true;
    }
}


//combination of R_AddDynamicLights and R_MarkLights
void R_StainNode (mnode_t *node, float *parms)
{
    mplane_t	*splitplane;
    float		dist;
    msurface_t	*surf;
    int			i;

    // node is a leaf
    if (node->contents < 0) return;

    splitplane = node->plane;
    dist = PlaneDiff((parms + 1), splitplane);

    if (dist > (*parms))
    {
        R_StainNode (node->children[0], parms);
        return;
    }

    if (dist < (-*parms))
    {
        R_StainNode (node->children[1], parms);
        return;
    }

    // mark the polygons
    surf = cl.worldmodel->surfaces + node->firstsurface;

    for (i = 0; i < node->numsurfaces; i++, surf++)
    {
        if (surf->flags & (SURF_DRAWSKY | SURF_DRAWTURB)) continue;

        R_StainSurf(surf, parms);
    }
    R_StainNode (node->children[0], parms);
    R_StainNode (node->children[1], parms);
}

void R_AddStain (float *org, float red, float green, float blue, float radius)
{
    float	parms[7];
    int		i;
    model_t *m;

    parms[0] = radius;
    parms[1] = org[0];
    parms[2] = org[1];
    parms[3] = org[2];

    parms[4] = red;
    parms[5] = green;
    parms[6] = blue;

    // worldmodel first
    R_StainNode (cl.worldmodel->nodes + cl.worldmodel->hulls[0].firstclipnode, parms);

    // now stain the rest of the models
    for (i = 1; i < MAX_MODELS; i++)
    {
        // get the current model
        m = cl.model_precache[i];

        // no more models
        if (!m) break;

        // don;t do * models (their data comes from the world)
        if (m->name[0] != '*') continue;

        // make sure we only get brush models
        if (m->type != mod_brush) continue;

        R_StainNode (m->nodes + m->hulls[0].firstclipnode, parms);
    }
}

