/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/


/*
============================================================================================================

								NEW MEMORY MANAGEMENT ROUTINES

	The original Hunk/Zone/Cache routines worked, but they were horribly complex, messy, and more suited
	to an OS without any built-in memory management of it's own (like DOS).  Time for them to die.

	The new "Heap" system lives alongside the old system for the time being, for the purpose of phased
	migration.  It works in a similar fashion to the old "Hunk" system, but is cleaner and more linear.

	Low allocations are freed at the start of each map.
	They may also be explicitly freed by using a Heap_LowMark/Heap_FreeToLowMark combo.
	High allocations are "scratch" memory and are only available until the next high allocation.

============================================================================================================
*/


#include "quakedef.h"

static byte		*heap_base = NULL;
static int		heap_low_mark = 0;
static int		heap_high_mark = 0;
static int		heap_size = 0;
volatile int	heap_checksum;


/*
=================
Heap_Check

Sanity checks the heap
=================
*/
static void Heap_Check (void)
{
    if (heap_low_mark >= heap_high_mark)
    {
        Sys_Error ("Heap_Check: fatal heap collision!  try increasing -heapsize");
    }

    if (heap_low_mark > heap_size)
    {
        Sys_Error ("Heap_Check: out of memory!");
    }
}


/*
=================
Heap_ScratchAlloc

Returns a chunk of "scratch" memory of the specified size from the top of the heap
=================
*/
void *Heap_ScratchAlloc (int size)
{
    byte *heap_ptr;

    // round off to 32 bytes
    size = (size + 31) & ~31;

    heap_ptr = &heap_base[heap_high_mark - size - 1];

    // temp
    heap_high_mark -= size;

    // sanity check
    Heap_Check ();

    // restore
    heap_high_mark += size;

    return heap_ptr;
}


/*
=================
Heap_ScratchZAlloc

Returns a chunk of "scratch" memory of the specified size from the top of the heap
Initialixes to 0 before returning
=================
*/
void *Heap_ScratchZAlloc (int size)
{
    byte *heap_ptr;

    // round off to 32 bytes
    size = (size + 31) & ~31;

    heap_ptr = &heap_base[heap_high_mark - size - 1];

    // temp
    heap_high_mark -= size;

    // sanity check
    Heap_Check ();

    // init to 0
    memset (heap_ptr, 0, size);

    // restore
    heap_high_mark += size;

    return heap_ptr;
}

/*
=================
Heap_LowMark

Returns the current low mark of the heap
=================
*/
int Heap_LowMark (void)
{
    return heap_low_mark;
}

/*
=================
Heap_FreeToLowMark

Sets the heap low mark to the specified new low mark
=================
*/
void Heap_FreeToLowMark (int mark)
{
    heap_low_mark = mark;
}

/*
=================
Heap_Alloc

Returns a chunk of memory of the specified size from the heap
=================
*/
void *Heap_Alloc (int size)
{
    byte *heap_ptr = &heap_base[heap_low_mark];

    // round off to 32 bytes
    size = (size + 31) & ~31;

    // increase the low mark
    heap_low_mark += size;

    // sanity check
    Heap_Check ();

    // return the buffer we just got
    return heap_ptr;
}

/*
=================
Heap_ZAlloc

Same as Heap_Alloc but initializes the memory to 0 before returning it
=================
*/
void *Heap_ZAlloc (int size)
{
    byte *heap_ptr = &heap_base[heap_low_mark];

    // round off to 32 bytes
    size = (size + 31) & ~31;

    // increase the low mark
    heap_low_mark += size;

    // sanity check
    Heap_Check ();

    // initialize to 0
    memset (heap_ptr, 0, size);

    // return the buffer we just got
    return heap_ptr;
}

/*
================
Heap_PageIn
================
*/
void Heap_PageIn (void *ptr, int size)
{
    byte	*x;
    int		m, n;

// touch all the memory to make sure it's there. The 16-page skip is to
// keep Win 95 from thinking we're trying to page ourselves in (we are
// doing that, of course, but there's no reason we shouldn't)
    x = (byte *)ptr;

    for (n=0 ; n<4 ; n++)
    {
        for (m = 0 ; m < (size - 16 * 0x1000) ; m += 4)
        {
            heap_checksum += *(int *)&x[m];
            heap_checksum += *(int *)&x[m + 16 * 0x1000];
        }
    }
}

/*
=================
Heap_MemAlloc

Standard windows memory interface
Can be used instead of malloc.
=================
*/
void *Heap_MemAlloc (size_t size)
{
    LPVOID	reserve;
    LPVOID	*buff;

    // if a null pointer was passed
    // get the fuck out of here !!!
    if (size < 0)
    {
        Sys_Error ("Heap_MemAlloc: bad size %i", size);
    }

    // reserve a chunk of memory the size the app called for
    reserve = VirtualAlloc(NULL, size, MEM_RESERVE, PAGE_NOACCESS);

    // bad puppy baaad
    if (!reserve)
    {
        Sys_Error ("Heap_MemAlloc: could not reserve %i MB of memory!", size / 1024 / 1024);
    }

    // allocate the reserved memory now that we made sure its ours
    buff = VirtualAlloc(reserve, size, MEM_COMMIT, PAGE_EXECUTE_READWRITE);

    // where did ma ram go boab !!!
    if (!buff)
    {
        Sys_Error ("Heap_MemAlloc: could not allocate %i MB of memory!", size / 1024 / 1024);
    }
    return buff;
}

/*
=================
Heap_MemAllocFree

Standard windows memory interface
Can be used instead of free.
=================
*/
void Heap_MemAllocFree (void *buffer)
{
    if (buffer)
    {
        VirtualFree(buffer, 0, MEM_RELEASE);
    }
}

/*
=================
Heap_Init

Initializes the heap system by just grabbing a chunk of memory.
Can be called multiple times if required to reinitialize the memory.
If called while a game is running, the game must be fully shut down.
=================
*/
void Heap_Init (int memsize)
{
    // free up the heap
    Heap_MemAllocFree (heap_base);

    // allocate the reserved memory now that we made sure its ours
    heap_base = (byte *)Heap_MemAlloc(memsize);

    // set the low mark
    heap_low_mark = 0;
    heap_size = heap_high_mark = memsize - 1;

    // sanity check
    Heap_Check ();

    Heap_PageIn (heap_base, memsize);
}



